# -*- coding: utf-8 -*-
"""

    mslib.mscolab._tests.utils
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Utility functions for mscolab

    This file is part of MSS.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2022 by the MSS team, see AUTHORS.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import fs
import os
import logging

from mslib.mscolab.conf import mscolab_settings


def get_recent_op_id(fm, user):
    operations = fm.list_operations(user)
    op_id = None
    if operations:
        op_id = operations[-1]["op_id"]
    return op_id


def get_session_id(sockets, u_id):
    s_id = None
    for ss in sockets:
        if ss["u_id"] == u_id:
            s_id = ss["s_id"]
    return s_id


def get_message_dict(message):
    return {
        "id": message.id,
        "u_id": message.u_id,
        "username": message.user.username,
        "text": message.text,
        "message_type": message.message_type,
        "reply_id": message.reply_id,
        "replies": [],
        "time": message.created_at.strftime("%Y-%m-%d, %H:%M:%S")
    }


def os_fs_create_dir(dir):
    if '://' in dir:
        try:
            _ = fs.open_fs(dir)
        except fs.errors.CreateFailed:
            logging.error(f'Make sure that the FS url "{dir}" exists')
        except fs.opener.errors.UnsupportedProtocol:
            logging.error(f'FS url "{dir}" not supported')
    else:
        _dir = os.path.expanduser(dir)
        if not os.path.exists(_dir):
            os.makedirs(_dir)


def create_files():
    os_fs_create_dir(mscolab_settings.MSCOLAB_DATA_DIR)
    os_fs_create_dir(mscolab_settings.UPLOAD_FOLDER)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_tableview_window.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_TableViewWindow(object):
    def setupUi(self, TableViewWindow):
        TableViewWindow.setObjectName("TableViewWindow")
        TableViewWindow.resize(1254, 472)
        TableViewWindow.setMinimumSize(QtCore.QSize(0, 0))
        self.centralwidget = QtWidgets.QWidget(TableViewWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tableWayPoints = QtWidgets.QTableView(self.centralwidget)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.tableWayPoints.setFont(font)
        self.tableWayPoints.setDragEnabled(True)
        self.tableWayPoints.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.tableWayPoints.setDefaultDropAction(QtCore.Qt.CopyAction)
        self.tableWayPoints.setAlternatingRowColors(True)
        self.tableWayPoints.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.tableWayPoints.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tableWayPoints.setObjectName("tableWayPoints")
        self.verticalLayout.addWidget(self.tableWayPoints)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.cbTools = QtWidgets.QComboBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbTools.sizePolicy().hasHeightForWidth())
        self.cbTools.setSizePolicy(sizePolicy)
        self.cbTools.setBaseSize(QtCore.QSize(0, 0))
        self.cbTools.setObjectName("cbTools")
        self.cbTools.addItem("")
        self.horizontalLayout.addWidget(self.cbTools)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.label = QtWidgets.QLabel(self.centralwidget)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.btAddWayPointToFlightTrack = QtWidgets.QPushButton(self.centralwidget)
        self.btAddWayPointToFlightTrack.setObjectName("btAddWayPointToFlightTrack")
        self.horizontalLayout.addWidget(self.btAddWayPointToFlightTrack)
        self.btCloneWaypoint = QtWidgets.QPushButton(self.centralwidget)
        self.btCloneWaypoint.setObjectName("btCloneWaypoint")
        self.horizontalLayout.addWidget(self.btCloneWaypoint)
        self.btDeleteWayPoint = QtWidgets.QPushButton(self.centralwidget)
        self.btDeleteWayPoint.setObjectName("btDeleteWayPoint")
        self.horizontalLayout.addWidget(self.btDeleteWayPoint)
        self.btInvertDirection = QtWidgets.QPushButton(self.centralwidget)
        self.btInvertDirection.setMinimumSize(QtCore.QSize(100, 0))
        self.btInvertDirection.setObjectName("btInvertDirection")
        self.horizontalLayout.addWidget(self.btInvertDirection)
        self.btRoundtrip = QtWidgets.QPushButton(self.centralwidget)
        self.btRoundtrip.setObjectName("btRoundtrip")
        self.horizontalLayout.addWidget(self.btRoundtrip)
        self.verticalLayout.addLayout(self.horizontalLayout)
        TableViewWindow.setCentralWidget(self.centralwidget)
        self.actionCloseWindow = QtWidgets.QAction(TableViewWindow)
        self.actionCloseWindow.setObjectName("actionCloseWindow")
        TableViewWindow.addAction(self.actionCloseWindow)

        self.retranslateUi(TableViewWindow)
        self.actionCloseWindow.triggered.connect(TableViewWindow.close)
        QtCore.QMetaObject.connectSlotsByName(TableViewWindow)

    def retranslateUi(self, TableViewWindow):
        _translate = QtCore.QCoreApplication.translate
        TableViewWindow.setWindowTitle(_translate("TableViewWindow", "Table View - Mission Support System"))
        self.cbTools.setItemText(0, _translate("TableViewWindow", "(select to open control)"))
        self.label.setText(_translate("TableViewWindow", "Waypoints:"))
        self.btAddWayPointToFlightTrack.setText(_translate("TableViewWindow", "insert"))
        self.btCloneWaypoint.setText(_translate("TableViewWindow", "clone"))
        self.btDeleteWayPoint.setText(_translate("TableViewWindow", "delete selected"))
        self.btInvertDirection.setText(_translate("TableViewWindow", "reverse"))
        self.btRoundtrip.setText(_translate("TableViewWindow", "make roundtrip"))
        self.actionCloseWindow.setText(_translate("TableViewWindow", "Close Window"))
        self.actionCloseWindow.setShortcut(_translate("TableViewWindow", "Ctrl+W"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_wms_dockwidget.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_WMSDockWidget(object):
    def setupUi(self, WMSDockWidget):
        WMSDockWidget.setObjectName("WMSDockWidget")
        WMSDockWidget.resize(756, 140)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(WMSDockWidget.sizePolicy().hasHeightForWidth())
        WMSDockWidget.setSizePolicy(sizePolicy)
        WMSDockWidget.setMinimumSize(QtCore.QSize(756, 0))
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(WMSDockWidget)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.cbLevel = QtWidgets.QComboBox(WMSDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbLevel.sizePolicy().hasHeightForWidth())
        self.cbLevel.setSizePolicy(sizePolicy)
        self.cbLevel.setMinimumSize(QtCore.QSize(180, 0))
        self.cbLevel.setMaximumSize(QtCore.QSize(300, 16777215))
        self.cbLevel.setObjectName("cbLevel")
        self.horizontalLayout_4.addWidget(self.cbLevel)
        self.tbLevel_back = QtWidgets.QToolButton(WMSDockWidget)
        self.tbLevel_back.setObjectName("tbLevel_back")
        self.horizontalLayout_4.addWidget(self.tbLevel_back)
        self.tbLevel_fwd = QtWidgets.QToolButton(WMSDockWidget)
        self.tbLevel_fwd.setObjectName("tbLevel_fwd")
        self.horizontalLayout_4.addWidget(self.tbLevel_fwd)
        spacerItem = QtWidgets.QSpacerItem(5, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem)
        self.cbAutoUpdate = QtWidgets.QCheckBox(WMSDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbAutoUpdate.sizePolicy().hasHeightForWidth())
        self.cbAutoUpdate.setSizePolicy(sizePolicy)
        self.cbAutoUpdate.setMaximumSize(QtCore.QSize(200, 16777215))
        self.cbAutoUpdate.setChecked(True)
        self.cbAutoUpdate.setObjectName("cbAutoUpdate")
        self.horizontalLayout_4.addWidget(self.cbAutoUpdate)
        self.cbTransparent = QtWidgets.QCheckBox(WMSDockWidget)
        self.cbTransparent.setChecked(True)
        self.cbTransparent.setObjectName("cbTransparent")
        self.horizontalLayout_4.addWidget(self.cbTransparent)
        self.cbCacheEnabled = QtWidgets.QCheckBox(WMSDockWidget)
        self.cbCacheEnabled.setChecked(True)
        self.cbCacheEnabled.setTristate(False)
        self.cbCacheEnabled.setObjectName("cbCacheEnabled")
        self.horizontalLayout_4.addWidget(self.cbCacheEnabled)
        self.btClearCache = QtWidgets.QPushButton(WMSDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.btClearCache.sizePolicy().hasHeightForWidth())
        self.btClearCache.setSizePolicy(sizePolicy)
        self.btClearCache.setMaximumSize(QtCore.QSize(100, 16777215))
        self.btClearCache.setObjectName("btClearCache")
        self.horizontalLayout_4.addWidget(self.btClearCache)
        self.gridLayout.addLayout(self.horizontalLayout_4, 2, 1, 1, 1)
        self.pbLayerSelect = QtWidgets.QPushButton(WMSDockWidget)
        self.pbLayerSelect.setObjectName("pbLayerSelect")
        self.gridLayout.addWidget(self.pbLayerSelect, 1, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(WMSDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 4, 0, 1, 1)
        self.label = QtWidgets.QLabel(WMSDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 2, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(WMSDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 3, 0, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.cbInitTime = QtWidgets.QComboBox(WMSDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbInitTime.sizePolicy().hasHeightForWidth())
        self.cbInitTime.setSizePolicy(sizePolicy)
        self.cbInitTime.setMinimumSize(QtCore.QSize(180, 0))
        self.cbInitTime.setMaximumSize(QtCore.QSize(300, 16777215))
        self.cbInitTime.setObjectName("cbInitTime")
        self.cbInitTime.addItem("")
        self.horizontalLayout_3.addWidget(self.cbInitTime)
        self.tbInitTime_cbback = QtWidgets.QToolButton(WMSDockWidget)
        self.tbInitTime_cbback.setObjectName("tbInitTime_cbback")
        self.horizontalLayout_3.addWidget(self.tbInitTime_cbback)
        self.tbInitTime_cbfwd = QtWidgets.QToolButton(WMSDockWidget)
        self.tbInitTime_cbfwd.setObjectName("tbInitTime_cbfwd")
        self.horizontalLayout_3.addWidget(self.tbInitTime_cbfwd)
        self.dteInitTime = QtWidgets.QDateTimeEdit(WMSDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.dteInitTime.sizePolicy().hasHeightForWidth())
        self.dteInitTime.setSizePolicy(sizePolicy)
        self.dteInitTime.setMinimumSize(QtCore.QSize(160, 0))
        self.dteInitTime.setMaximumSize(QtCore.QSize(300, 16777215))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        font.setStrikeOut(False)
        self.dteInitTime.setFont(font)
        self.dteInitTime.setCorrectionMode(QtWidgets.QAbstractSpinBox.CorrectToPreviousValue)
        self.dteInitTime.setCalendarPopup(False)
        self.dteInitTime.setTimeSpec(QtCore.Qt.UTC)
        self.dteInitTime.setObjectName("dteInitTime")
        self.horizontalLayout_3.addWidget(self.dteInitTime)
        self.tbInitTime_back = QtWidgets.QToolButton(WMSDockWidget)
        self.tbInitTime_back.setObjectName("tbInitTime_back")
        self.horizontalLayout_3.addWidget(self.tbInitTime_back)
        self.cbInitTime_step = QtWidgets.QComboBox(WMSDockWidget)
        self.cbInitTime_step.setObjectName("cbInitTime_step")
        self.horizontalLayout_3.addWidget(self.cbInitTime_step)
        self.tbInitTime_fwd = QtWidgets.QToolButton(WMSDockWidget)
        self.tbInitTime_fwd.setObjectName("tbInitTime_fwd")
        self.horizontalLayout_3.addWidget(self.tbInitTime_fwd)
        self.gridLayout.addLayout(self.horizontalLayout_3, 3, 1, 1, 1)
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.cbValidTime = QtWidgets.QComboBox(WMSDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbValidTime.sizePolicy().hasHeightForWidth())
        self.cbValidTime.setSizePolicy(sizePolicy)
        self.cbValidTime.setMinimumSize(QtCore.QSize(180, 0))
        self.cbValidTime.setMaximumSize(QtCore.QSize(300, 16777215))
        self.cbValidTime.setObjectName("cbValidTime")
        self.horizontalLayout_8.addWidget(self.cbValidTime)
        self.tbValidTime_cbback = QtWidgets.QToolButton(WMSDockWidget)
        self.tbValidTime_cbback.setArrowType(QtCore.Qt.NoArrow)
        self.tbValidTime_cbback.setObjectName("tbValidTime_cbback")
        self.horizontalLayout_8.addWidget(self.tbValidTime_cbback)
        self.tbValidTime_cbfwd = QtWidgets.QToolButton(WMSDockWidget)
        self.tbValidTime_cbfwd.setObjectName("tbValidTime_cbfwd")
        self.horizontalLayout_8.addWidget(self.tbValidTime_cbfwd)
        self.dteValidTime = QtWidgets.QDateTimeEdit(WMSDockWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.dteValidTime.sizePolicy().hasHeightForWidth())
        self.dteValidTime.setSizePolicy(sizePolicy)
        self.dteValidTime.setMinimumSize(QtCore.QSize(160, 0))
        self.dteValidTime.setMaximumSize(QtCore.QSize(300, 16777215))
        self.dteValidTime.setDate(QtCore.QDate(2009, 12, 25))
        self.dteValidTime.setMinimumDateTime(QtCore.QDateTime(QtCore.QDate(1999, 12, 31), QtCore.QTime(8, 0, 0)))
        self.dteValidTime.setMinimumDate(QtCore.QDate(1999, 12, 31))
        self.dteValidTime.setMinimumTime(QtCore.QTime(8, 0, 0))
        self.dteValidTime.setCalendarPopup(False)
        self.dteValidTime.setTimeSpec(QtCore.Qt.UTC)
        self.dteValidTime.setObjectName("dteValidTime")
        self.horizontalLayout_8.addWidget(self.dteValidTime)
        self.tbValidTime_back = QtWidgets.QToolButton(WMSDockWidget)
        self.tbValidTime_back.setObjectName("tbValidTime_back")
        self.horizontalLayout_8.addWidget(self.tbValidTime_back)
        self.cbValidTime_step = QtWidgets.QComboBox(WMSDockWidget)
        self.cbValidTime_step.setObjectName("cbValidTime_step")
        self.horizontalLayout_8.addWidget(self.cbValidTime_step)
        self.tbValidTime_fwd = QtWidgets.QToolButton(WMSDockWidget)
        self.tbValidTime_fwd.setObjectName("tbValidTime_fwd")
        self.horizontalLayout_8.addWidget(self.tbValidTime_fwd)
        self.gridLayout.addLayout(self.horizontalLayout_8, 4, 1, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.lLayerName = QtWidgets.QLabel(WMSDockWidget)
        self.lLayerName.setWordWrap(False)
        self.lLayerName.setObjectName("lLayerName")
        self.horizontalLayout.addWidget(self.lLayerName)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.line = QtWidgets.QFrame(WMSDockWidget)
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.horizontalLayout.addWidget(self.line)
        self.btClearMap = QtWidgets.QPushButton(WMSDockWidget)
        self.btClearMap.setMaximumSize(QtCore.QSize(200, 16777215))
        self.btClearMap.setObjectName("btClearMap")
        self.horizontalLayout.addWidget(self.btClearMap)
        self.btGetMap = QtWidgets.QPushButton(WMSDockWidget)
        self.btGetMap.setMaximumSize(QtCore.QSize(200, 16777215))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.btGetMap.setFont(font)
        self.btGetMap.setObjectName("btGetMap")
        self.horizontalLayout.addWidget(self.btGetMap)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 1, 1, 1)
        self.verticalLayout_4.addLayout(self.gridLayout)

        self.retranslateUi(WMSDockWidget)
        QtCore.QMetaObject.connectSlotsByName(WMSDockWidget)

    def retranslateUi(self, WMSDockWidget):
        _translate = QtCore.QCoreApplication.translate
        WMSDockWidget.setWindowTitle(_translate("WMSDockWidget", "WMS Dock Widget"))
        self.cbLevel.setToolTip(_translate("WMSDockWidget", "Elevation values provided by the WMS server."))
        self.tbLevel_back.setToolTip(_translate("WMSDockWidget", "Previous Level (Down)"))
        self.tbLevel_back.setText(_translate("WMSDockWidget", "<"))
        self.tbLevel_back.setShortcut(_translate("WMSDockWidget", "Down"))
        self.tbLevel_fwd.setToolTip(_translate("WMSDockWidget", "Next Level (Up)"))
        self.tbLevel_fwd.setText(_translate("WMSDockWidget", ">"))
        self.tbLevel_fwd.setShortcut(_translate("WMSDockWidget", "Up"))
        self.cbAutoUpdate.setToolTip(_translate("WMSDockWidget", "Automatically request an updated map when the layer parameters have changed."))
        self.cbAutoUpdate.setText(_translate("WMSDockWidget", "Auto-Update"))
        self.cbTransparent.setText(_translate("WMSDockWidget", "Transparent"))
        self.cbCacheEnabled.setToolTip(_translate("WMSDockWidget", "Enable the image cache (retrieved images will be stored locally to speed up repeated retrievals)."))
        self.cbCacheEnabled.setText(_translate("WMSDockWidget", "Use Cache"))
        self.btClearCache.setToolTip(_translate("WMSDockWidget", "Clear all cache contents."))
        self.btClearCache.setText(_translate("WMSDockWidget", "Clear Cache"))
        self.pbLayerSelect.setText(_translate("WMSDockWidget", "Server/Layer"))
        self.label_3.setText(_translate("WMSDockWidget", "Valid:"))
        self.label.setText(_translate("WMSDockWidget", "Level:"))
        self.label_2.setText(_translate("WMSDockWidget", "Initialisation:"))
        self.cbInitTime.setToolTip(_translate("WMSDockWidget", "Forecast initialisation time (base time) values provided by the WMS server."))
        self.cbInitTime.setItemText(0, _translate("WMSDockWidget", "2010-12-12T00:00:00Z"))
        self.tbInitTime_cbback.setToolTip(_translate("WMSDockWidget", "Previous Init-Time (Alt+Left)"))
        self.tbInitTime_cbback.setText(_translate("WMSDockWidget", "<"))
        self.tbInitTime_cbback.setShortcut(_translate("WMSDockWidget", "Alt+Left"))
        self.tbInitTime_cbfwd.setToolTip(_translate("WMSDockWidget", "Next Init-Time (Alt+Right)"))
        self.tbInitTime_cbfwd.setText(_translate("WMSDockWidget", ">"))
        self.tbInitTime_cbfwd.setShortcut(_translate("WMSDockWidget", "Alt+Right"))
        self.dteInitTime.setToolTip(_translate("WMSDockWidget", "You can also specify an initialisation date here."))
        self.dteInitTime.setDisplayFormat(_translate("WMSDockWidget", "yyyy/MM/dd hh:mm UTC"))
        self.tbInitTime_back.setText(_translate("WMSDockWidget", "<"))
        self.tbInitTime_fwd.setText(_translate("WMSDockWidget", ">"))
        self.cbValidTime.setToolTip(_translate("WMSDockWidget", "Valid time values provided by the WMS server."))
        self.tbValidTime_cbback.setToolTip(_translate("WMSDockWidget", "Previous Valid-Time (Left)"))
        self.tbValidTime_cbback.setText(_translate("WMSDockWidget", "<"))
        self.tbValidTime_cbback.setShortcut(_translate("WMSDockWidget", "Left"))
        self.tbValidTime_cbfwd.setToolTip(_translate("WMSDockWidget", "Next Valid-Time (Right)"))
        self.tbValidTime_cbfwd.setText(_translate("WMSDockWidget", ">"))
        self.tbValidTime_cbfwd.setShortcut(_translate("WMSDockWidget", "Right"))
        self.dteValidTime.setToolTip(_translate("WMSDockWidget", "Specify the time value here, especially if the server does not provide predefined values. Keep in mind that the specified value may not be available from the server, though."))
        self.dteValidTime.setDisplayFormat(_translate("WMSDockWidget", "yyyy/MM/dd hh:mm UTC"))
        self.tbValidTime_back.setText(_translate("WMSDockWidget", "<"))
        self.tbValidTime_fwd.setText(_translate("WMSDockWidget", ">"))
        self.lLayerName.setText(_translate("WMSDockWidget", "No Layer selected"))
        self.btClearMap.setText(_translate("WMSDockWidget", "remove"))
        self.btGetMap.setToolTip(_translate("WMSDockWidget", "Request the selected layer(s) (Return)"))
        self.btGetMap.setText(_translate("WMSDockWidget", "retrieve"))
        self.btGetMap.setShortcut(_translate("WMSDockWidget", "Return"))



# -*- coding: utf-8 -*-
"""

    mslib.msui.airdata_dockwidget
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Control to load airports and airspaces into the top view.

    This file is part of MSS.

    :copyright: Copyright 2021 May Bär
    :copyright: Copyright 2021-2022 by the MSS team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import pycountry
from mslib.utils.qt import ui_airdata_dockwidget as ui
from PyQt5 import QtWidgets, QtCore
from mslib.utils.config import save_settings_qsettings, load_settings_qsettings
from mslib.utils.airdata import get_available_airspaces, update_airspace, get_airports


class AirdataDockwidget(QtWidgets.QWidget, ui.Ui_AirdataDockwidget):
    def __init__(self, parent=None, view=None):
        super(AirdataDockwidget, self).__init__(parent)
        self.setupUi(self)
        self.view = view
        self.view.redrawn.connect(self.redraw_map)

        code_to_name = {country.alpha_2.lower(): country.name for country in pycountry.countries}
        self.cbAirspaces.addItems([f"{code_to_name.get(airspace[0].split('_')[0], 'Unknown')} "
                                   f"{airspace[0].split('_')[0]}"
                                   for airspace in get_available_airspaces()])
        self.cbAirportType.addItems(["small_airport", "medium_airport", "large_airport", "heliport", "balloonport",
                                     "seaplane_base", "closed"])
        self.cbAirportType.empty_text = "Click here to select airports..."
        self.cbAirspaces.empty_text = "Click here to select airspaces..."

        self.settings_tag = "airdatadock"
        settings = load_settings_qsettings(self.settings_tag, {"draw_airports": False, "draw_airspaces": False,
                                                               "airspaces": [], "airport_type": [],
                                                               "filter_airspaces": False, "filter_from": 0,
                                                               "filter_to": 100})

        self.btDownload.clicked.connect(lambda: get_airports(True))
        self.btDownloadAsp.clicked.connect(lambda: update_airspace(True, [airspace.split(" ")[-1] for airspace in
                                                                          self.cbAirspaces.currentData()]))
        self.btApply.clicked.connect(self.redraw_map)

        self.cbAirspaces.currentTextChanged.connect(self.adjust_ui_airspaces)
        self.cbAirportType.currentTextChanged.connect(self.adjust_ui_airports)

        self.cbDrawAirports.setChecked(settings["draw_airports"])
        self.cbDrawAirspaces.setChecked(settings["draw_airspaces"])

        for airspace in settings["airspaces"]:
            i = self.cbAirspaces.findText(airspace)
            if i != -1:
                self.cbAirspaces.model().item(i).setCheckState(QtCore.Qt.Checked)
        for airport in settings["airport_type"]:
            i = self.cbAirportType.findText(airport)
            if i != -1:
                self.cbAirportType.model().item(i).setCheckState(QtCore.Qt.Checked)
        self.cbAirspaces.updateText()
        self.cbFilterAirspaces.setChecked(settings["filter_airspaces"])
        self.sbFrom.setValue(settings["filter_from"])
        self.sbTo.setValue(settings["filter_to"])

    def adjust_ui_airspaces(self):
        """
        Disables and unchecks, or vice versa, airspace UI elements depending on the current user selection
        """
        airspaces_enabled = len(self.cbAirspaces.currentData()) > 0
        self.cbDrawAirspaces.setChecked(airspaces_enabled)
        self.cbDrawAirspaces.setEnabled(airspaces_enabled)
        self.btDownloadAsp.setEnabled(airspaces_enabled)

    def adjust_ui_airports(self):
        """
        Disables and unchecks, or vice versa, airport UI elements depending on the current user selection
        """
        airports_enabled = len(self.cbAirportType.currentData()) > 0
        self.cbDrawAirports.setChecked(airports_enabled)
        self.cbDrawAirports.setEnabled(airports_enabled)
        self.btDownload.setEnabled(airports_enabled)

    def redraw_map(self):
        if self.view.map is not None:
            self.view.map.set_draw_airports(self.cbDrawAirports.isChecked(), port_type=self.cbAirportType.currentData())
            self.view.map.set_draw_airspaces(self.cbDrawAirspaces.isChecked(), self.cbAirspaces.currentData(),
                                             (self.sbFrom.value(), self.sbTo.value())
                                             if self.cbFilterAirspaces.isChecked() else None)
            self.view.draw()
        self.save_settings()

    def save_settings(self):
        settings_dict = {
            "draw_airports": self.cbDrawAirports.isChecked(),
            "airport_type": self.cbAirportType.currentData(),
            "draw_airspaces": self.cbDrawAirspaces.isChecked(),
            "airspaces": self.cbAirspaces.currentData(),
            "filter_airspaces": self.cbFilterAirspaces.isChecked(),
            "filter_from": self.sbFrom.value(),
            "filter_to": self.sbTo.value(),
        }
        save_settings_qsettings(self.settings_tag, settings_dict)



# -*- coding: utf-8 -*-
"""

    mslib.msui.flighttrack
    ~~~~~~~~~~~~~~~~~~~~~~

    Data model representing a flight track. The model is derived from
    QAbstractTableModel, so that it can directly be connected to any Qt view.

    For better understanding of the code, compare to the 'ships' example
    from chapter 14/16 of 'Rapid GUI Programming with Python and Qt: The
    Definitive Guide to PyQt Programming' (Mark Summerfield).

    The model includes a method for computing the distance between waypoints
    and for the entire flight track.

    This file is part of MSS.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2022 by the MSS team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import datetime
import logging
import os

import fs
import xml.dom.minidom
import xml.parsers.expat

from PyQt5 import QtGui, QtCore, QtWidgets

from mslib import __version__
from mslib.utils.units import units
from mslib.utils.coordinate import find_location, path_points, get_distance
from mslib.utils import thermolib
from mslib.utils.config import config_loader, save_settings_qsettings, load_settings_qsettings
from mslib.utils.config import MSUIDefaultConfig as mss_default
from mslib.utils.qt import variant_to_string, variant_to_float
from mslib.msui.performance_settings import DEFAULT_PERFORMANCE

from mslib.utils import writexml
xml.dom.minidom.Element.writexml = writexml
# Constants for identifying the table columns when the WaypointsTableModel is
# used with a QTableWidget.
LOCATION, LAT, LON, FLIGHTLEVEL, PRESSURE = list(range(5))
TIME_UTC = 9


def seconds_to_string(seconds):
    """
    Format a time given in seconds to a string HH:MM:SS. Used for the
    'leg time/cum. time' columns of the table view.
    """
    hours, seconds = divmod(int(seconds), 3600)
    minutes, seconds = divmod(seconds, 60)
    return f"{hours:02d}:{minutes:02d}:{seconds:02d}"


TABLE_FULL = [
    ("Location                   ", lambda waypoint: waypoint.location, True),
    ("Lat\n(+-90)", lambda waypoint: round(float(waypoint.lat), 2), True),
    ("Lon\n(+-180)", lambda waypoint: round(float(waypoint.lon), 2), True),
    ("Flightlevel", lambda waypoint: waypoint.flightlevel, True),
    ("Pressure\n(hPa)", lambda waypoint: QtCore.QLocale().toString(waypoint.pressure / 100., 'f', 2), True),
    ("Leg dist.\n(km [nm])", lambda waypoint: f"{int(waypoint.distance_to_prev):d} "
                                              f"[{int(waypoint.distance_to_prev / 1.852):d}]", False),
    ("Cum. dist.\n(km [nm])", lambda waypoint: f"{int(waypoint.distance_total):d} "
                                               f"[{int(waypoint.distance_total / 1.852):d}]", False),
    ("Leg time", lambda waypoint: seconds_to_string(waypoint.leg_time), False),
    ("Cum. time", lambda waypoint: seconds_to_string(waypoint.cum_time), False),
    ("Time (UTC)", lambda waypoint: waypoint.utc_time.strftime("%Y-%m-%d %H:%M:%S"), False),
    ("Rem. fuel\n(lb)", lambda waypoint: f"{int(waypoint.rem_fuel):d}", False),
    ("Aircraft\nweight (lb)", lambda waypoint: f"{int(waypoint.weight):d}", False),
    ("Ceiling\naltitude (hft)", lambda waypoint: f"{waypoint.ceiling_alt:d}", False),
    ("Ascent rate\n(ft/minute)", lambda waypoint: f"{waypoint.ascent_rate:d}", False),
    ("Comments                        ", lambda waypoint: waypoint.comments, True),
]

TABLE_SHORT = [TABLE_FULL[_i] for _i in range(7)] + [TABLE_FULL[-1]] + [("", lambda _: "", False)] * 8


class Waypoint(object):
    """
    Represents a waypoint with position, altitude and further
    properties. Used internally by WaypointsTableModel.
    """

    def __init__(self, lat=0, lon=0, flightlevel=0, location="", comments=""):
        self.location = location
        locations = config_loader(dataset='locations')
        if location in locations:
            self.lat, self.lon = locations[location]
        else:
            self.lat = lat
            self.lon = lon
        self.flightlevel = flightlevel
        self.pressure = thermolib.flightlevel2pressure(flightlevel * units.hft).magnitude
        self.distance_to_prev = 0.
        self.distance_total = 0.
        self.comments = comments

        # Performance fields (for values read from the flight performance
        # service).
        self.leg_time = None  # time from previous waypoint
        self.cum_time = None  # total time of flight
        self.utc_time = None  # time in UTC since given takeoff time
        self.leg_fuel = None  # fuel consumption since previous waypoint
        self.rem_fuel = None  # total fuel consumption
        self.weight = None  # aircraft gross weight
        self.ceiling_alt = None  # aircraft ceiling altitude
        self.ascent_rate = None  # aircraft ascent rate

        self.wpnumber_major = None
        self.wpnumber_minor = None

    def __str__(self):
        """
        String representation of the waypoint (e.g., when used with the print
        statement).
        """
        return f"WAYPOINT(LAT={self.lat:f}, LON={self.lon:f}, FL={self.flightlevel:f})"


class WaypointsTableModel(QtCore.QAbstractTableModel):
    """
    Qt-QAbstractTableModel-derived data structure representing a flight
    track composed of a number of waypoints.

    Objects of this class can be directly connected to any Qt view that is
    able to handle tables models.

    Provides methods to store and load the model to/from an XML file, to compute
    distances between the individual waypoints, and to interpret the results of
    flight performance calculations.
    """

    def __init__(self, name="", filename=None, waypoints=None, mscolab_mode=False, data_dir=mss_default.mss_dir,
                 xml_content=None):
        super(WaypointsTableModel, self).__init__()
        self.name = name  # a name for this flight track
        self.filename = filename  # filename for store/load
        self.data_dir = data_dir
        self.modified = False  # for "save on exit"
        self.waypoints = []  # user-defined waypoints
        # file-save events are handled in a different manner
        self.mscolab_mode = mscolab_mode

        # self.aircraft.setErrorHandling("permissive")
        self.settings_tag = "performance"
        self.load_settings()

        # If a filename is passed to the constructor, load data from this file.
        if filename is not None:
            self.load_from_ftml(filename)

        # If xml string is passed to constructor, load data from that
        elif xml_content is not None:
            self.load_from_xml_data(xml_content)

        if waypoints:
            self.replace_waypoints(waypoints)

    def load_settings(self):
        """
        Load settings from the file self.settingsfile.
        """
        self.performance_settings = load_settings_qsettings(self.settings_tag, DEFAULT_PERFORMANCE)

    def save_settings(self):
        """
        Save the current settings (map appearance) to the file
        self.settingsfile.
        """
        save_settings_qsettings(self.settings_tag, self.performance_settings)

    def flags(self, index):
        """
        Used to specify which table columns can be edited by the user;
        overrides the corresponding QAbstractTableModel method.
        """
        if not index.isValid():
            return QtCore.Qt.ItemIsEnabled
        column = index.column()
        table = TABLE_SHORT
        if self.performance_settings["visible"]:
            table = TABLE_FULL
        if table[column][2]:
            return QtCore.Qt.ItemFlags(
                int(QtCore.QAbstractTableModel.flags(self, index) |
                    QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsDragEnabled |
                    QtCore.Qt.ItemIsDropEnabled))
        else:
            return QtCore.Qt.ItemFlags(
                int(QtCore.QAbstractTableModel.flags(self, index) |
                    QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled))

    def data(self, index, role=QtCore.Qt.DisplayRole):
        """
        Return a data field at the given index (of type QModelIndex,
        specifying row and column); overrides the corresponding
        QAbstractTableModel method.

        NOTE: Other roles (e.g. for display appearance) could be specified in
        this method as well. Cf. the 'ships' example in chapter 14/16 of 'Rapid
        GUI Programming with Python and Qt: The Definitive Guide to PyQt
        Programming' (Mark Summerfield).
        """
        waypoints = self.waypoints

        if not index.isValid() or not (0 <= index.row() < len(waypoints)):
            return QtCore.QVariant()
        waypoint = waypoints[index.row()]
        column = index.column()
        if role == QtCore.Qt.DisplayRole:
            if self.performance_settings["visible"]:
                return QtCore.QVariant(TABLE_FULL[column][1](waypoint))
            else:
                return QtCore.QVariant(TABLE_SHORT[column][1](waypoint))
        elif role == QtCore.Qt.TextAlignmentRole:
            return QtCore.QVariant(int(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter))
        return QtCore.QVariant()

    def waypoint_data(self, row):
        """
        Get the waypoint object defining the given row.
        """
        return self.waypoints[row]

    def all_waypoint_data(self):
        """
        Return the entire list of waypoints.
        """
        return self.waypoints

    def intermediate_points(self, numpoints=101, connection="greatcircle"):
        """
        Compute intermediate points between the waypoints.

        See mss_util.path_points() for additional arguments.

        Returns lats, lons.
        """
        return path_points(
            [wp.lat for wp in self.waypoints],
            [wp.lon for wp in self.waypoints],
            times=[wp.utc_time for wp in self.waypoints],
            numpoints=numpoints, connection=connection)

    def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
        """
        Return data describing the table header; overrides the
        corresponding QAbstractTableModel method.
        """
        if role == QtCore.Qt.TextAlignmentRole:
            if orientation == QtCore.Qt.Horizontal:
                return QtCore.QVariant(int(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter))
            return QtCore.QVariant(int(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter))
        if role != QtCore.Qt.DisplayRole:
            return QtCore.QVariant()
        # Return the names of the table columns.
        if orientation == QtCore.Qt.Horizontal:
            if self.performance_settings["visible"]:
                return QtCore.QVariant(TABLE_FULL[section][0])
            else:
                return QtCore.QVariant(TABLE_SHORT[section][0])
        # Table rows (waypoints) are labelled with their number (= number of
        # waypoint).
        return QtCore.QVariant(int(section))

    def rowCount(self, index=QtCore.QModelIndex()):
        """
        Number of waypoints in the model.
        """
        return len(self.waypoints)

    def columnCount(self, index=QtCore.QModelIndex()):
        return len(TABLE_FULL)

    def setData(self, index, value, role=QtCore.Qt.EditRole, update=True):
        """
        Change a data element of the flight track; overrides the
        corresponding QAbstractTableModel method.

        NOTE: Performance computations loose their validity if a change is made.
        """
        if index.isValid() and 0 <= index.row() < len(self.waypoints):
            waypoint = self.waypoints[index.row()]
            column = index.column()
            index2 = index  # in most cases only one field is being changed
            if column == LOCATION:
                waypoint.location = variant_to_string(value)
            elif column == LAT:
                try:
                    # The table fields accept basically any input.
                    # If the string cannot be converted to "float" (raises ValueError), the user input is discarded.
                    value = variant_to_float(value)
                except TypeError as ex:
                    logging.error("unexpected error: %s %s %s %s", type(ex), ex, type(value), value)
                except ValueError as ex:
                    logging.error("%s", ex)
                else:
                    waypoint.lat = value
                    waypoint.location = ""
                    loc = find_location(waypoint.lat, waypoint.lon, 1e-3)
                    if loc is not None:
                        waypoint.lat, waypoint.lon = loc[0]
                        waypoint.location = loc[1]
                    # A change of position requires an update of the distances.
                    if update:
                        self.update_distances(index.row())
                    # Notify the views that items between the edited item and
                    # the distance item of the corresponding waypoint have been
                    # changed.
                    # Delete the location name -- it won't be valid anymore
                    # after its coordinates have been changed.
                    index2 = self.createIndex(index.row(), LOCATION)
            elif column == LON:
                try:
                    # The table fields accept basically any input.
                    # If the string cannot be converted to "float" (raises ValueError), the user input is discarded.
                    value = variant_to_float(value)
                except TypeError as ex:
                    logging.error("unexpected error: %s %s %s %s", type(ex), ex, type(value), value)
                except ValueError as ex:
                    logging.error("%s", ex)
                else:
                    waypoint.lon = value
                    waypoint.location = ""
                    loc = find_location(waypoint.lat, waypoint.lon, 1e-3)
                    if loc is not None:
                        waypoint.lat, waypoint.lon = loc[0]
                        waypoint.location = loc[1]
                    if update:
                        self.update_distances(index.row())
                    index2 = self.createIndex(index.row(), LOCATION)
            elif column == FLIGHTLEVEL:
                try:
                    # The table fields accept basically any input.
                    # If the string cannot be converted to "float" (raises ValueError), the user input is discarded.
                    flightlevel = variant_to_float(value)
                    pressure = float(thermolib.flightlevel2pressure(flightlevel * units.hft).magnitude)
                except TypeError as ex:
                    logging.error("unexpected error: %s %s %s %s", type(ex), ex, type(value), value)
                except ValueError as ex:
                    logging.error("%s", ex)
                else:
                    waypoint.flightlevel = flightlevel
                    waypoint.pressure = pressure
                    if update:
                        self.update_distances(index.row())
                    # need to notify view of the second item that has been
                    # changed as well.
                    index2 = self.createIndex(index.row(), PRESSURE)
            elif column == PRESSURE:
                try:
                    # The table fields accept basically any input.
                    # If the string cannot be converted to "float" (raises ValueError), the user input is discarded.
                    pressure = variant_to_float(value) * 100  # convert hPa to Pa
                    if pressure > 200000:
                        raise ValueError
                    flightlevel = float(round(thermolib.pressure2flightlevel(pressure * units.Pa).magnitude))
                    pressure = float(thermolib.flightlevel2pressure(flightlevel * units.hft).magnitude)
                except TypeError as ex:
                    logging.error("unexpected error: %s %s %s %s", type(ex), ex, type(value), value)
                except ValueError as ex:
                    logging.error("%s", ex)
                else:
                    waypoint.pressure = pressure
                    waypoint.flightlevel = flightlevel
                    if update:
                        self.update_distances(index.row())
                    index2 = self.createIndex(index.row(), FLIGHTLEVEL)
            else:
                waypoint.comments = variant_to_string(value)
            self.modified = True
            # Performance computations loose their validity if a change is made.
            if update:
                self.dataChanged.emit(index, index2)
            return True
        return False

    def insertRows(self, position, rows=1, index=QtCore.QModelIndex(),
                   waypoints=None):
        """
        Insert waypoint; overrides the corresponding QAbstractTableModel
        method.
        """
        if not waypoints:
            waypoints = [Waypoint(0, 0, 0)] * rows

        assert len(waypoints) == rows, (waypoints, rows)

        self.beginInsertRows(QtCore.QModelIndex(), position,
                             position + rows - 1)
        for row, wp in enumerate(waypoints):
            self.waypoints.insert(position + row, wp)

        self.update_distances(position, rows=rows)
        self.endInsertRows()
        self.modified = True
        return True

    def removeRows(self, position, rows=1, index=QtCore.QModelIndex()):
        """
        Remove waypoint; overrides the corresponding QAbstractTableModel
        method.
        """
        # beginRemoveRows emits rowsAboutToBeRemoved(index, first, last).
        self.beginRemoveRows(QtCore.QModelIndex(), position,
                             position + rows - 1)
        self.waypoints = self.waypoints[:position] + self.waypoints[position + rows:]
        if position < len(self.waypoints):
            self.update_distances(position, rows=min(rows, len(self.waypoints) - position))

        # endRemoveRows emits rowsRemoved(index, first, last).
        self.endRemoveRows()
        self.modified = True
        return True

    def update_distances(self, position, rows=1):
        """
        Update all distances in a flight track that are affected by a
        waypoint change involving <rows> waypoints starting at index
        <position>.

        Distances are computed along great circles.

        If rows=1, the distance to the previous waypoint is updated for
        waypoints <position> and <position+1>. The total flight track distance
        is updated for all waypoint following <position>.

        If rows>1, the distances to the previous waypoints are updated
        according to the number of modified waypoints.
        """
        waypoints = self.waypoints
        aircraft = self.performance_settings["aircraft"]

        def get_duration_fuel(flightlevel0, flightlevel1, distance, weight, lastleg):
            if flightlevel0 == flightlevel1:
                tas, fuelflow = aircraft.get_cruise_performance(flightlevel0 * 100, weight)
                duration = 3600. * distance / (1.852 * tas)  # convert to s (tas is in nm/h)
                leg_fuel = duration * fuelflow / 3600.
                return duration, leg_fuel
            else:
                if flightlevel0 < flightlevel1:
                    duration0, dist0, fuel0 = aircraft.get_climb_performance(flightlevel0 * 100, weight)
                    duration1, dist1, fuel1 = aircraft.get_climb_performance(flightlevel1 * 100, weight)
                else:
                    duration0, dist0, fuel0 = aircraft.get_descent_performance(flightlevel0 * 100, weight)
                    duration1, dist1, fuel1 = aircraft.get_descent_performance(flightlevel1 * 100, weight)
                duration = (duration1 - duration0) * 60  # convert from min to s
                dist = (dist1 - dist0) * 1.852  # convert from nm to km
                fuel = fuel1 - fuel0
                if lastleg:
                    duration_p, fuel_p = get_duration_fuel(flightlevel0, flightlevel0, distance - dist, weight, False)
                else:
                    duration_p, fuel_p = get_duration_fuel(flightlevel1, flightlevel1, distance - dist, weight, False)
                return duration + duration_p, fuel + fuel_p

        pos = position
        for offset in range(rows):
            pos = position + offset
            wp1 = waypoints[pos]
            # The distance to the first waypoint is zero.
            if pos == 0:
                wp1.distance_to_prev = 0.
                wp1.distance_total = 0.

                wp1.leg_time = 0  # time from previous waypoint
                wp1.cum_time = 0  # total time of flight
                wp1.utc_time = self.performance_settings["takeoff_time"].toPyDateTime()
                wp1.weight = self.performance_settings["takeoff_weight"]
                wp1.leg_fuel = 0
                wp1.rem_fuel = self.performance_settings["takeoff_weight"] - self.performance_settings["empty_weight"]
                wp1.ascent_rate = 0
            else:
                wp0 = waypoints[pos - 1]
                wp1.distance_to_prev = get_distance(
                    wp0.lat, wp0.lon, wp1.lat, wp1.lon)

                last = (pos - 1 == rows)
                time, fuel = get_duration_fuel(
                    wp0.flightlevel, wp1.flightlevel, wp1.distance_to_prev, wp0.weight, lastleg=last)
                wp1.leg_time = time
                wp1.cum_time = wp0.cum_time + wp1.leg_time
                wp1.utc_time = wp0.utc_time + datetime.timedelta(seconds=wp1.leg_time)
                wp1.leg_fuel = fuel
                wp1.rem_fuel = wp0.rem_fuel - wp1.leg_fuel
                wp1.weight = wp0.weight - wp1.leg_fuel
                if wp1.leg_time != 0:
                    wp1.ascent_rate = int((wp1.flightlevel - wp0.flightlevel) * 100 / (wp1.leg_time / 60))
                else:
                    wp1.ascent_rate = 0
            wp1.ceiling_alt = aircraft.get_ceiling_altitude(wp1.weight)

        # Update the distance of the following waypoint as well.
        if pos < len(waypoints) - 1:
            wp2 = waypoints[pos + 1]
            wp2.distance_to_prev = get_distance(
                wp1.lat, wp1.lon, wp2.lat, wp2.lon)
            if wp2.leg_time != 0:
                wp2.ascent_rate = int((wp2.flightlevel - wp1.flightlevel) * 100 / (wp2.leg_time / 60))
            else:
                wp2.ascent_rate = 0

        # Update total distances of waypoint at index position and all
        # following waypoints.
        for i in range(max(min(position, 1), 1), len(waypoints)):
            wp0 = waypoints[i - 1]
            wp1 = waypoints[i]
            wp1.distance_total = wp0.distance_total + wp1.distance_to_prev
            wp1.weight = wp0.weight - wp0.leg_fuel
            last = (i + 1 == len(waypoints))
            time, fuel = get_duration_fuel(
                wp0.flightlevel, wp1.flightlevel, wp1.distance_to_prev, wp0.weight, lastleg=last)

            wp1.leg_time = time
            wp1.cum_time = wp0.cum_time + wp1.leg_time
            wp1.utc_time = wp0.utc_time + datetime.timedelta(seconds=wp1.leg_time)
            wp1.leg_fuel = fuel
            wp1.rem_fuel = wp0.rem_fuel - wp1.leg_fuel
            wp1.weight = wp0.weight - wp1.leg_fuel
            wp1.ceiling_alt = aircraft.get_ceiling_altitude(wp1.weight)

        index1 = self.createIndex(0, TIME_UTC)
        self.dataChanged.emit(index1, index1)

    def invert_direction(self):
        self.waypoints = self.waypoints[::-1]
        if len(self.waypoints) > 0:
            self.waypoints[0].distance_to_prev = 0
            self.waypoints[0].distance_total = 0
        for i in range(1, len(self.waypoints)):
            wp_comm = self.waypoints[i].comments
            if len(wp_comm) == 9 and wp_comm.startswith("Hexagon "):
                wp_comm = f"Hexagon {(8 - int(wp_comm[-1])):d}"
                self.waypoints[i].comments = wp_comm
        self.update_distances(position=0, rows=len(self.waypoints))
        index = self.index(0, 0)

        self.layoutChanged.emit()
        self.dataChanged.emit(index, index)

    def replace_waypoints(self, new_waypoints):
        self.waypoints = []
        self.insertRows(0, rows=len(new_waypoints), waypoints=new_waypoints)

    def save_to_ftml(self, filename=None):
        """
        Save the flight track to an XML file.

        Arguments:
        filename -- complete path to the file to save. If None, a previously
                    specified filename will be used. If no filename has been
                    specified at all, a ValueError exception will be raised.
        """
        if not filename:
            raise ValueError("filename to save flight track cannot be None or empty")

        self.filename = filename
        self.name = fs.path.basename(filename.replace(".ftml", "").strip())
        doc = self.get_xml_doc()
        dirname, name = fs.path.split(self.filename)
        file_dir = fs.open_fs(dirname)
        with file_dir.open(name, 'w') as file_object:
            doc.writexml(file_object, indent="  ", addindent="  ", newl="\n", encoding="utf-8")
        file_dir.close()

    def get_xml_doc(self):
        doc = xml.dom.minidom.Document()
        ft_el = doc.createElement("FlightTrack")
        ft_el.setAttribute("version", __version__)
        doc.appendChild(ft_el)
        # The list of waypoint elements.
        wp_el = doc.createElement("ListOfWaypoints")
        ft_el.appendChild(wp_el)
        for wp in self.waypoints:
            element = doc.createElement("Waypoint")
            wp_el.appendChild(element)
            element.setAttribute("location", str(wp.location))
            element.setAttribute("lat", str(wp.lat))
            element.setAttribute("lon", str(wp.lon))
            element.setAttribute("flightlevel", str(wp.flightlevel))
            comments = doc.createElement("Comments")
            comments.appendChild(doc.createTextNode(str(wp.comments)))
            element.appendChild(comments)
        return doc

    def get_xml_content(self):
        doc = self.get_xml_doc()
        return doc.toprettyxml(indent="  ", newl="\n")

    def load_from_ftml(self, filename):
        """
        Load a flight track from an XML file at <filename>.
        """
        _dirname, _name = os.path.split(filename)
        _fs = fs.open_fs(_dirname)
        xml_content = _fs.readtext(_name)
        name = os.path.basename(filename.replace(".ftml", "").strip())
        self.load_from_xml_data(xml_content, name)

    def load_from_xml_data(self, xml_content, name="Flight track"):
        try:
            doc = xml.dom.minidom.parseString(xml_content)
        except xml.parsers.expat.ExpatError as ex:
            raise SyntaxError(str(ex))

        ft_el = doc.getElementsByTagName("FlightTrack")[0]

        self.name = name

        waypoints_list = []
        for wp_el in ft_el.getElementsByTagName("Waypoint"):

            location = wp_el.getAttribute("location")
            lat = float(wp_el.getAttribute("lat"))
            lon = float(wp_el.getAttribute("lon"))
            flightlevel = float(wp_el.getAttribute("flightlevel"))
            comments = wp_el.getElementsByTagName("Comments")[0]
            # If num of comments is 0(null comment), then return ''
            if len(comments.childNodes):
                comments = comments.childNodes[0].data.strip()
            else:
                comments = str('')

            waypoints_list.append(Waypoint(lat, lon, flightlevel,
                                           location=location,
                                           comments=comments))
        self.replace_waypoints(waypoints_list)

    def get_filename(self):
        return self.filename


#
# CLASS  WaypointDelegate
#


class WaypointDelegate(QtWidgets.QItemDelegate):
    """
    Qt delegate class for the appearance of the table view. Based on the
    'ships' example in chapter 14/16 of 'Rapid GUI Programming with Python
    and Qt: The Definitive Guide to PyQt Programming' (Mark Summerfield).
    """

    def __init__(self, parent=None):
        super(WaypointDelegate, self).__init__(parent)

    def paint(self, painter, option, index):
        """
        Colors waypoints with a minor waypoint number (i.e. intermediate
        waypoints generated by the flight performance service) in red.
        """
        wpnumber_minor = index.model().waypoint_data(index.row()).wpnumber_minor
        if wpnumber_minor is not None and wpnumber_minor > 0:
            newpalette = QtGui.QPalette(option.palette)
            colour = QtGui.QColor(170, 0, 0)  # dark red
            newpalette.setColor(QtGui.QPalette.Text, colour)
            colour = QtGui.QColor(255, 255, 0)  # yellow
            newpalette.setColor(QtGui.QPalette.HighlightedText, colour)
            option.palette = newpalette
        QtWidgets.QItemDelegate.paint(self, painter, option, index)

    def createEditor(self, parent, option, index):
        """
        Create a combobox listing predefined locations in the LOCATION
        column.
        """
        if index.column() == LOCATION:
            combobox = QtWidgets.QComboBox(parent)
            locations = config_loader(dataset='locations')
            adds = list(locations.keys())
            if self.parent() is not None:
                for loc in [wp.location for wp in self.parent().waypoints_model.all_waypoint_data() if
                            wp.location != ""]:
                    if loc not in adds:
                        adds.append(loc)
            combobox.addItems(sorted(adds))

            combobox.setEditable(True)
            return combobox
        else:
            # All other columns get the standard editor.
            return QtWidgets.QItemDelegate.createEditor(self, parent, option, index)

    def setEditorData(self, editor, index):
        text = index.model().data(index, QtCore.Qt.DisplayRole).value()
        if index.column() in (LOCATION,):
            i = editor.findText(text)
            if i == -1:
                i = 0
            editor.setCurrentIndex(i)
        else:
            QtWidgets.QItemDelegate.setEditorData(self, editor, index)

    def setModelData(self, editor, model, index):
        """
        For the LOCATION column: If the user selects a location from the
        combobox, get the corresponding coordinates.
        """
        if index.column() == LOCATION:
            loc = editor.currentText()
            locations = config_loader(dataset='locations')
            if loc in locations:
                lat, lon = locations[loc]
                # Don't update distances and flight performance twice, hence
                # set update=False for LAT.
                model.setData(index.sibling(index.row(), LAT), QtCore.QVariant(lat), update=False)
                model.setData(index.sibling(index.row(), LON), QtCore.QVariant(lon))
            else:
                for wp in self.parent().waypoints_model.all_waypoint_data():
                    if loc == wp.location:
                        lat, lon = wp.lat, wp.lon
                        # Don't update distances and flight performance twice, hence
                        # set update=False for LAT.
                        model.setData(index.sibling(index.row(), LAT), QtCore.QVariant(lat), update=False)
                        model.setData(index.sibling(index.row(), LON), QtCore.QVariant(lon))

            model.setData(index, QtCore.QVariant(editor.currentText()))
        else:
            QtWidgets.QItemDelegate.setModelData(self, editor, model, index)



# -*- coding: utf-8 -*-
"""

    mslib.utils.config
    ~~~~~~~~~~~~~~~~

    Collection of functions all around config handling.

    This file is part of MSS.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2022 by the MSS team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import sys
from PyQt5 import QtCore

import copy
import json
import logging
import fs
import os
import tempfile

from mslib.utils import FatalUserError
from mslib.msui import constants
from mslib.support.qt_json_view.datatypes import match_type, UrlType, StrType


class MSUIDefaultConfig(object):
    """Central configuration for the Mission Support System User Interface
       Application (msui).

    DESCRIPTION:
    ============

    This file includes configuration settings central to the entire
    Mission Support User Interface (msui). Among others, define
     -- available map projections
     -- vertical section interpolation options
     -- the lists of predefined web service URLs
     -- predefined waypoints for the table view
    in this file.

    Do not change any value for good reasons.
    Your values can be set in your personal msui_settings.json file
    """
    # this skips the verification of the user token on each mscolab request
    mscolab_skip_verify_user_token = True

    # Default for general filepicker. Pick "default", "qt", or "fs"
    filepicker_default = "default"

    # dir where msui output files are stored
    data_dir = "~/mssdata"

    # layout of different views, with immutable they can't resized
    layout = {"topview": [963, 702],
              "sideview": [913, 557],
              "linearview": [913, 557],
              "tableview": [1236, 424],
              "immutable": False}

    # Predefined map regions to be listed in the corresponding topview combobox
    predefined_map_sections = {
        "01 Europe (cyl)": {"CRS": "EPSG:4326",
                            "map": {"llcrnrlon": -15.0, "llcrnrlat": 35.0,
                                    "urcrnrlon": 30.0, "urcrnrlat": 65.0}},
        "02 Germany (cyl)": {"CRS": "EPSG:4326",
                             "map": {"llcrnrlon": 5.0, "llcrnrlat": 45.0,
                                     "urcrnrlon": 15.0, "urcrnrlat": 57.0}},
        "03 Global (cyl)": {"CRS": "EPSG:4326",
                            "map": {"llcrnrlon": -180.0, "llcrnrlat": -90.0,
                                    "urcrnrlon": 180.0, "urcrnrlat": 90.0}},
        "04 Northern Hemisphere (stereo)": {"CRS": "MSS:stere,0,90,90",
                                            "map": {"llcrnrlon": -45.0, "llcrnrlat": 0.0,
                                                    "urcrnrlon": 135.0, "urcrnrlat": 0.0}}
    }

    # Side View.
    # The following two parameters are passed to the WMS in the BBOX
    # argument when a vertical cross section is requested.

    # Number of interpolation points used to interpolate the flight track
    # to a great circle.
    num_interpolation_points = 201

    # Number of x-axis labels in the side view.
    num_labels = 10

    # Web Map Service Client.
    # Settings for the WMS client. Set the URLs of WMS servers that appear
    # by default in the WMS control (for examples, see
    # http://external.opengis.org/twiki_public/bin/view/MetOceanDWG/MetocWMS_Servers).
    # Also set the location of the image file cache and its size.

    # URLs of default WMS servers.
    default_WMS = [
        "http://localhost:8081/",
        "http://eumetview.eumetsat.int/geoserver/wms",
        "https://apps.ecmwf.int/wms/?token=public"
    ]

    default_VSEC_WMS = [
        "http://localhost:8081/"
    ]

    default_LSEC_WMS = [
        "http://localhost:8081/"
    ]

    # URLs of default mscolab servers
    default_MSCOLAB = [
        "http://localhost:8083",
    ]

    # mail address to sign in
    MSCOLAB_mailid = ""

    # password to sign in
    MSCOLAB_password = ""

    # category for MSC operations
    MSCOLAB_category = "default"

    # dictionary of MSC servers {"http://www.your-mscolab-server.de" : ("youruser", "yourpassword")}
    MSC_login = {}

    # timeout of Url request
    WMS_request_timeout = 30

    WMS_preload = []

    # dictionary of WMS servers {"http://www.your-wms-server.de" : ("youruser", "yourpassword")}
    WMS_login = {}

    # WMS image cache settings:
    wms_cache = os.path.join(tempfile.gettempdir(), "msui_wms_cache")

    # Maximum size of the cache in bytes.
    wms_cache_max_size_bytes = 20 * 1024 * 1024

    # Maximum age of a cached file in seconds.
    wms_cache_max_age_seconds = 5 * 86400

    wms_prefetch = {
        "validtime_fwd": 0,
        "validtime_bck": 0,
        "level_up": 0,
        "level_down": 0
    }

    locations = {
        "EDMO": [48.08, 11.28],
        "Hannover": [52.37, 9.74],
        "Hamburg": [53.55, 9.99],
        "Juelich": [50.92, 6.36],
        "Leipzig": [51.34, 12.37],
        "Muenchen": [48.14, 11.57],
        "Stuttgart": [48.78, 9.18],
        "Wien": [48.20833, 16.373064],
        "Zugspitze": [47.42, 10.98],
        "Kiruna": [67.821, 20.336],
        "Ny-Alesund": [78.928, 11.986],
        "Zhukovsky": [55.6, 38.116],
        "Paphos": [34.775, 32.425],
        "Sharjah": [25.35, 55.65],
        "Brindisi": [40.658, 17.947],
        "Nagpur": [21.15, 79.083],
        "Mumbai": [19.089, 72.868],
        "Delhi": [28.566, 77.103],
    }

    # Main application: Template for new flight tracks
    # Flight track template that is used when a new flight track is
    # created. Specify a list of place names that can be found in the
    # "locations" dictionary defined above.
    new_flighttrack_template = ["Nagpur", "Delhi"]

    # This configures the flight level for waypoints inserted by the
    # flighttrack template
    new_flighttrack_flightlevel = 0

    # None is not wanted here
    proxies = {}

    # ToDo configurable later
    # mscolab server
    mscolab_server_url = "http://localhost:8083"
    # ToDo refactor to rename this to data_dir/mss_data_dir
    # mss dir
    mss_dir = "~/mss"

    # list of gravatar email ids to automatically fetch
    gravatar_ids = []

    # dictionary for export plugins, e.g.  {"Text": ["txt", "mslib.plugins.io.text", "save_to_txt"] }
    export_plugins = {}

    # dictionary for import plugins, e.g. { "FliteStar": ["txt", "mslib.plugins.io.flitestar", "load_from_flitestar"] }
    import_plugins = {}

    # dictionary to make title, label and ticklabel sizes for topview and sideview configurable.
    # You can put your default value here, whatever you want to give,it should be a number.
    topview = {"plot_title_size": 10,
               "axes_label_size": 10}

    sideview = {"plot_title_size": 10,
                "axes_label_size": 10}

    linearview = {"plot_title_size": 10,
                  "axes_label_size": 10}

    # Dictionary options with fixed key/value pairs
    fixed_dict_options = ["layout", "wms_prefetch", "topview", "sideview", "linearview"]

    # Fixed key/value pair options
    key_value_options = [
        'mscolab_skip_verify_user_token',
        'filepicker_default',
        'mss_dir',
        'data_dir',
        'num_labels',
        'num_interpolation_points',
        'new_flighttrack_flightlevel',
        'MSCOLAB_mailid',
        'MSCOLAB_password',
        'MSCOLAB_category',
        'mscolab_server_url',
        'wms_cache',
        'wms_cache_max_size_bytes',
        'wms_cache_max_age_seconds',
        'WMS_request_timeout',
    ]

    # Dictionary options with predefined structure
    dict_option_structure = {
        "predefined_map_sections": {
            "new_map_section": {
                "CRS": "crs_value",
                "map": {
                    "llcrnrlon": 0.0,
                    "llcrnrlat": 0.0,
                    "urcrnrlon": 0.0,
                    "urcrnrlat": 0.0,
                },
            }
        },
        "MSC_login": {
            "http://www.your-mscolab-server.de": ["yourusername", "yourpassword"],
        },
        "WMS_login": {
            "http://www.your-wms-server.de": ["yourusername", "yourpassword"],
        },
        "locations": {
            "new-location": [0.0, 0.0],
        },
        "export_plugins": {
            "plugin-name": ["extension", "module", "function", "default"],
        },
        "import_plugins": {
            "plugin-name": ["extension", "module", "function", "default"],
        },
        "proxies": {
            "https": "https://proxy.com",
        },
    }

    # List options with predefined structure
    list_option_structure = {
        "default_WMS": ["https://wms-server-url.com"],
        "default_VSEC_WMS": ["https://vsec-wms-server-url.com"],
        "default_LSEC_WMS": ["https://lsec-wms-server-url.com"],
        "default_MSCOLAB": ["https://mscolab-server-url.com"],
        "new_flighttrack_template": ["new-location"],
        "gravatar_ids": ["example@email.com"],
        "WMS_preload": ["https://wms-preload-url.com"],
    }

    config_descriptions = {
        "filepicker_default": "Documentation Required",
        "data_dir": "Documentation Required",
        "predefined_map_sections": "Documentation Required",
        "num_interpolation_points": "Documentation Required",
        "num_labels": "Documentation Required",
        "default_WMS": "Documentation Required",
        "default_VSEC_WMS": "Documentation Required",
        "default_LSEC_WMS": "Documentation Required",
        "default_MSCOLAB": "Documentation Required",
        "MSCOLAB_mailid": "Documentation Required",
        "MSCOLAB_password": "Documentation Required",
        "MSC_login": "Documentation Required",
        "WMS_request_timeout": "Documentation Required",
        "WMS_preload": "Documentation Required",
        "WMS_login": "Documentation Required",
        "wms_cache": "Documentation Required",
        "wms_cache_max_size_bytes": "Documentation Required",
        "wms_cache_max_age_seconds": "Documentation Required",
        "wms_prefetch": "Documentation Required",
        "locations": "Documentation Required",
        "new_flighttrack_template": "Documentation Required",
        "new_flighttrack_flightlevel": "Documentation Required",
        "proxies": "Documentation Required",
        "mscolab_server_url": "Documentation Required",
        "mss_dir": "Documentation Required",
        "gravatar_ids": "Documentation Required",
        "export_plugins": "Documentation Required",
        "import_plugins": "Documentation Required",
        "layout": "Documentation Required",
        "topview": "Documentation Required",
        "sideview": "Documentation Required",
        "linearview": "Documentation Required",
    }


# default options as dictionary
default_options = dict(MSUIDefaultConfig.__dict__)
for key in [
    "__module__",
    "__doc__",
    "__dict__",
    "__weakref__",
    "fixed_dict_options",
    "dict_option_structure",
    "list_option_structure",
    "key_value_options",
    "config_descriptions",
]:
    del default_options[key]


# user options as dictionary
user_options = copy.deepcopy(default_options)


def read_config_file(path=constants.MSUI_SETTINGS):
    """
    reads a config file and updates global user_options

    Args:
        path: path of config file

    Note:
        sole purpose of the path argument is to be able to test with example config files
    """
    path = path.replace("\\", "/")
    dir_name, file_name = fs.path.split(path)
    json_file_data = {}
    with fs.open_fs(dir_name) as _fs:
        if _fs.exists(file_name):
            file_content = _fs.readtext(file_name)
            try:
                json_file_data = json.loads(file_content, object_pairs_hook=dict_raise_on_duplicates_empty)
            except json.JSONDecodeError as e:
                logging.error(f"Error while loading json file {e}")
                error_message = f"Unexpected error while loading config\n{e}"
                raise FatalUserError(error_message)
            except ValueError as e:
                logging.error(f"Error while loading json file {e}")
                error_message = f"Invalid keys detected in config\n{e}"
                raise FatalUserError(error_message)
        else:
            error_message = f"MSS config File '{path}' not found"
            raise FileNotFoundError(error_message)

    global user_options
    if json_file_data:
        user_options = merge_dict(copy.deepcopy(default_options), json_file_data)
        logging.debug("Merged default and user settings")
    else:
        user_options = copy.deepcopy(default_options)
        logging.debug("No user settings found, using default settings")


def modify_config_file(data, path=constants.MSUI_SETTINGS):
    """
    modifies a config file

    Args:
        data: data to be modified/written
        path: path of config file

    Note:
        sole purpose of the path argument is to be able to test with example config files
    """
    path = path.replace("\\", "/")
    dir_name, file_name = fs.path.split(path)
    json_file_data = {}
    with fs.open_fs(dir_name) as _fs:
        if _fs.exists(file_name):
            try:
                file_content = _fs.readtext(file_name)
                json_file_data = json.loads(file_content, object_pairs_hook=dict_raise_on_duplicates_empty)
                json_file_data_copy = copy.deepcopy(json_file_data)
                for key in data:
                    if key not in json_file_data:
                        json_file_data_copy[key] = config_loader(dataset=key, default=True)
                modified_data = merge_dict(json_file_data_copy, data)
                logging.debug("Merged default and user settings")
                _fs.writetext(file_name, json.dumps(modified_data, indent=4))
                read_config_file()
            except json.JSONDecodeError as e:
                logging.error(f"Error while loading json file {e}")
                error_message = f"Unexpected error while loading config\n{e}"
                raise FatalUserError(error_message)
            except ValueError as e:
                logging.error(f"Error while loading json file {e}")
                error_message = f"Invalid keys detected in config\n{e}"
                raise FatalUserError(error_message)
        else:
            error_message = f"MSS config File '{path}' not found"
            raise FileNotFoundError(error_message)


def config_loader(dataset=None, default=False):
    """
    Function for returning config value

    Args:
        dataset: section to pull from json file
        default: option to return default config for the dataset

    Returns: a the dataset value or the config as dictionary

    """
    if dataset is not None and dataset not in user_options:
        raise KeyError(f"requested dataset '{dataset}' not in defaults!")

    if dataset is not None:
        if default:
            return default_options[dataset]
        return user_options[dataset]
    else:
        if default:
            return default_options
        return user_options


def save_settings_qsettings(tag, settings, ignore_test=False):
    """
    Saves a dictionary settings to disk.

    :param tag: string specifying the settings
    :param settings: dictionary of settings
    :return: None
    """
    assert isinstance(tag, str)
    assert isinstance(settings, dict)
    if not ignore_test and "pytest" in sys.modules:
        return settings

    q_settings = QtCore.QSettings("msui", "msui-core")
    file_path = q_settings.fileName()
    logging.debug("storing settings for %s to %s", tag, file_path)
    try:
        q_settings.setValue(tag, QtCore.QVariant(settings))
    except (OSError, IOError) as ex:
        logging.warning("Problems storing %s settings (%s: %s).", tag, type(ex), ex)
    return settings


def load_settings_qsettings(tag, default_settings=None, ignore_test=False):
    """
    Loads a dictionary of settings from disk. May supply a dictionary of default settings
    to return in case the settings file is not present or damaged. The default_settings one will
    be updated by the restored one so one may rely on all keys of the default_settings dictionary
    being present in the returned dictionary.

    :param tag: string specifying the settings
    :param default_settings: dictionary of settings or None
    :return: dictionary of settings
    """
    if default_settings is None:
        default_settings = {}
    assert isinstance(default_settings, dict)
    if not ignore_test and "pytest" in sys.modules:
        return default_settings

    settings = {}
    q_settings = QtCore.QSettings("msui", "msui-core")
    file_path = q_settings.fileName()
    logging.debug("loading settings for %s from %s", tag, file_path)
    try:
        settings = q_settings.value(tag)
    except Exception as ex:
        logging.error("Problems reloading stored %s settings (%s: %s). Switching to default",
                      tag, type(ex), ex)
    if isinstance(settings, dict):
        default_settings.update(settings)
    return default_settings


def merge_dict(existing_dict, new_dict):
    """
    Merge two dictionaries by comparing all the options from
    the MSUIDefaultConfig class

    Arguments:
    existing_dict -- Dict to merge new_dict into
    new_dict -- Dict with new values
    """
    # Check if dictionary options with fixed key/value pairs match data types from default
    for key in MSUIDefaultConfig.fixed_dict_options:
        if key in new_dict:
            existing_dict[key] = compare_data(
                existing_dict[key], new_dict[key]
            )[0]

    # Check if dictionary options with predefined structure match data types from default
    dos = copy.deepcopy(MSUIDefaultConfig.dict_option_structure)
    # adding plugin structure : ["extension", "module", "function"] to
    # recognize user plugin options that don't have the optional filepicker option set
    dos["import_plugins"]["plugin-name-a"] = dos["import_plugins"]["plugin-name"][:3]
    dos["export_plugins"]["plugin-name-a"] = dos["export_plugins"]["plugin-name"][:3]
    for key in dos:
        if key in new_dict:
            temp_data = {}
            for option_key in new_dict[key]:
                for dos_key_key in dos[key]:
                    data, match = compare_data(dos[key][dos_key_key], new_dict[key][option_key])
                    if match:
                        temp_data[option_key] = new_dict[key][option_key]
                        break
            if temp_data != {}:
                existing_dict[key] = temp_data

    # Check if list options with predefined structure match data types from default
    los = copy.deepcopy(MSUIDefaultConfig.list_option_structure)
    for key in los:
        if key in new_dict:
            temp_data = []
            for i in range(len(new_dict[key])):
                for los_key_item in los[key]:
                    data, match = compare_data(los_key_item, new_dict[key][i])
                    if match:
                        temp_data.append(data)
                        break
            if temp_data != []:
                existing_dict[key] = temp_data

    # Check if options with fixed key/value pair structure match data types from default
    for key in MSUIDefaultConfig.key_value_options:
        if key in new_dict:
            data, match = compare_data(existing_dict[key], new_dict[key])
            if match:
                existing_dict[key] = data

    # add filepicker default to import and export plugins if missing
    for plugin_type in ["import_plugins", "export_plugins"]:
        if plugin_type in existing_dict:
            for plugin in existing_dict[plugin_type]:
                if len(existing_dict[plugin_type][plugin]) == 3:
                    existing_dict[plugin_type][plugin].append(
                        existing_dict.get("filepicker_default", "default")
                    )

    return existing_dict


def compare_data(default, user_data):
    """
    Recursively compares two dictionaries based on qt_json_view datatypes
    and returns default or user_data appropriately.

    Arguments:
    default -- Dict to return if datatype not matching
    user_data -- Dict to return if datatype is matching
    """
    # If data is neither list not dict type, compare individual type
    if not isinstance(default, dict) and not isinstance(default, list):
        if isinstance(default, float) and isinstance(user_data, int):
            user_data = float(default)
        if isinstance(match_type(default), UrlType) and isinstance(match_type(user_data), StrType):
            return user_data, True
        if isinstance(match_type(default), type(match_type(user_data))):
            return user_data, True
        else:
            return default, False

    data = copy.deepcopy(default)
    matches = []
    # If data is list type, compare all values in list
    if isinstance(default, list) and isinstance(user_data, list):
        if len(default) == len(user_data):
            for i in range(len(default)):
                data[i], match = compare_data(default[i], user_data[i])
                matches.append(match)
        else:
            return default, False

    # If data is dict type, goes through the dict and update
    elif isinstance(default, dict) and isinstance(user_data, dict):
        if default.keys() == user_data.keys():
            for key in default:
                if key in user_data:
                    data[key], match = compare_data(default[key], user_data[key])
                    matches.append(match)
                else:
                    matches.append(False)
        else:
            return default, False

    return data, all(matches)


def dict_raise_on_duplicates_empty(ordered_pairs):
    """Reject duplicate and empty keys."""
    accepted = {}
    for key, value in ordered_pairs:
        if key in accepted:
            raise ValueError(f"duplicate key found: {key}")
        elif key == "":
            raise ValueError("empty key found")
        else:
            accepted[key] = value
    return accepted



# -*- coding: utf-8 -*-
"""

    mslib.mscolab.conf.py.example
    ~~~~~~~~~~~~~~~~~~~~

    config for mscolab.

    This file is part of MSS.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2022 by the MSS team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import logging
import secrets

try:
    from mscolab_settings import mscolab_settings
    logging.info("Using user defined settings")
except ImportError as ex:
    logging.warning(u"Couldn't import mscolab_settings (ImportError:'%s'), using dummy config.", ex)

    class mscolab_settings(object):
        import os
        import logging

        # expire token in seconds
        # EXPIRATION = 86400

        # Which origins are allowed to communicate with your server
        CORS_ORIGINS = ["*"]

        # dir where msui output files are stored
        BASE_DIR = os.path.expanduser("~")

        DATA_DIR = os.path.join(BASE_DIR, "colabdata")

        # mscolab data directory
        MSCOLAB_DATA_DIR = os.path.join(DATA_DIR, 'filedata')

        # MYSQL CONNECTION STRING: "mysql+pymysql://<username>:<password>@<host>:<port>/<db_name>?charset=utf8mb4"
        SQLALCHEMY_DB_URI = 'sqlite:///' + os.path.join(DATA_DIR, 'mscolab.db')

        # mscolab file upload settings
        UPLOAD_FOLDER = os.path.join(DATA_DIR, 'uploads')
        MAX_UPLOAD_SIZE = 2 * 1024 * 1024  # 2MB

        # used to generate and parse tokens
        SECRET_KEY = secrets.token_urlsafe(16)

        # used to generate the password token
        SECURITY_PASSWORD_SALT = secrets.token_urlsafe(16)

        STUB_CODE = """<?xml version="1.0" encoding="utf-8"?>
        <FlightTrack version="1.7.6">
          <ListOfWaypoints>
            <Waypoint flightlevel="250" lat="67.821" location="Kiruna" lon="20.336">
              <Comments></Comments>
            </Waypoint>
            <Waypoint flightlevel="250" lat="78.928" location="Ny-Alesund" lon="11.986">
              <Comments></Comments>
            </Waypoint>
          </ListOfWaypoints>
        </FlightTrack>
        """
        enable_basic_http_authentication = False

        # enable verification by Mail
        USER_VERIFICATION = False

        # mail settings
        # MAIL_SERVER = 'localhost'
        # MAIL_PORT = 25
        # MAIL_USE_TLS = False
        # MAIL_USE_SSL = True

        # mail authentication
        # MAIL_USERNAME = os.environ.get('APP_MAIL_USERNAME')
        # MAIL_PASSWORD = os.environ.get('APP_MAIL_PASSWORD')

        # mail accounts
        # MAIL_DEFAULT_SENDER = 'MSS@localhost'



# -*- coding: utf-8 -*-
"""

    mslib.msui.mscolab_operation
    ~~~~~~~~~~~~~~~~~~~~~

    Mscolab operation window, to display chat, file change

    This file is part of MSS.

    :copyright: 2019 Shivashis Padhi
    :copyright: Copyright 2019-2022 by the MSS team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import datetime
import json

import fs
import requests
from markdown import Markdown
from markdown.extensions import Extension
from werkzeug.urls import url_join

from mslib.mscolab.models import MessageType
from PyQt5 import QtCore, QtGui, QtWidgets
from mslib.utils.qt import get_open_filename, get_save_filename, show_popup
from mslib.msui.qt5 import ui_mscolab_operation_window as ui
from mslib.utils.config import config_loader


# We need to override the KeyPressEvent in QTextEdit to disable the default behaviour of enter key.
class MessageTextEdit(QtWidgets.QTextEdit):
    def keyPressEvent(self, event):
        # Shift + Enter for new line
        if event.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter) and event.modifiers() & QtCore.Qt.ShiftModifier:
            cursor = self.textCursor()
            cursor.insertText("\n")
            self.setTextCursor(cursor)
            return
        # Only Enter for send/edit message
        elif event.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
            chat_window = self.parent().parent().parent()
            if chat_window.active_edit_id is None:
                chat_window.send_message()
            else:
                chat_window.edit_message()
            return
        super().keyPressEvent(event)


class MSColabChatWindow(QtWidgets.QMainWindow, ui.Ui_MscolabOperation):
    """Derives QMainWindow to provide some common functionality to all
       MSUI view windows.
    """
    name = "MSColab Chat Window"
    identifier = None

    viewCloses = QtCore.pyqtSignal(name="viewCloses")
    reloadWindows = QtCore.pyqtSignal(name="reloadWindows")

    def __init__(self, token, op_id, user, operation_name, access_level, conn, parent=None,
                 mscolab_server_url=config_loader(dataset="default_MSCOLAB")):
        """
        token: access_token
        op_id: operation id
        user: logged in user
        operation_name: active operation name,
        access_level: access level of user logged in
        conn: to send messages, recv messages, if a direct slot-signal can't be setup
            to be connected at parents'
        parent: widget parent
        mscolab_server_url: server url for mscolab
        """
        super(MSColabChatWindow, self).__init__(parent)
        self.setupUi(self)

        self.mscolab_server_url = mscolab_server_url
        self.token = token
        self.user = user
        self.op_id = op_id
        self.operation_name = operation_name
        self.conn = conn
        self.access_level = access_level
        self.text = ""
        self.attachment = None
        self.attachment_type = None
        self.active_edit_id = None
        self.active_message_reply = None
        self.current_search_index = None
        self.markdown = Markdown(extensions=['nl2br', 'sane_lists', DeregisterSyntax()])
        self.messageText = MessageTextEdit(self.centralwidget)
        self.setup_message_text()
        # Signals
        self.searchMessageLineEdit.textChanged.connect(self.handle_search_text_changed)
        self.searchPrevBtn.clicked.connect(self.handle_prev_message_search)
        self.searchNextBtn.clicked.connect(self.handle_next_message_search)
        self.previewBtn.clicked.connect(self.toggle_preview)
        self.sendMessageBtn.clicked.connect(self.send_message)
        self.uploadBtn.clicked.connect(self.handle_upload)
        self.editMessageBtn.clicked.connect(self.edit_message)
        self.cancelBtn.clicked.connect(self.send_message_state)
        # Socket Connection handlers
        self.conn.signal_operation_permissions_updated.connect(self.handle_permissions_updated)
        self.conn.signal_message_receive.connect(self.handle_incoming_message)
        self.conn.signal_message_reply_receive.connect(self.handle_incoming_message_reply)
        self.conn.signal_message_edited.connect(self.handle_message_edited)
        self.conn.signal_message_deleted.connect(self.handle_deleted_message)
        # Set Label text
        self.set_label_text()
        # Hide Edit Message section
        self.send_message_state()
        # load all users
        self.load_users()
        # load messages
        self.load_all_messages()
        if access_level == "viewer":
            self.messageText.setEnabled(False)
            self.previewBtn.setEnabled(False)
            self.uploadBtn.setEnabled(False)
            self.sendMessageBtn.setEnabled(False)

    # UI SET UP METHODS
    def setup_message_text(self):
        self.messageText.setAcceptRichText(False)
        self.messageText.setTextInteractionFlags(
            QtCore.Qt.LinksAccessibleByKeyboard | QtCore.Qt.LinksAccessibleByMouse | QtCore.Qt.TextBrowserInteraction |
            QtCore.Qt.TextEditable | QtCore.Qt.TextEditorInteraction | QtCore.Qt.TextSelectableByKeyboard |
            QtCore.Qt.TextSelectableByMouse)
        self.messageText.setPlaceholderText("Enter message here.\nPress enter to send.\nShift+Enter to add a new line.")
        self.messageText.setObjectName("messageText")
        vbox_layout = QtWidgets.QVBoxLayout()
        vbox_layout.addWidget(self.messageText)
        vbox_layout.setSpacing(0)
        vbox_layout.setContentsMargins(0, 0, 0, 0)
        self.messageTextContainer.setLayout(vbox_layout)

    def set_label_text(self):
        self.user_info.setText(f"Logged in: {self.user['username']}")
        self.proj_info.setText(f"Operation: {self.operation_name}")

    def get_img_dimensions(self, image):
        # TODO: CHECK WHY SCROLL BAR COMES?
        max_height = self.messageText.size().height()
        height = max_height - 4
        width = image.width() * height / image.height()
        return width, height

    def display_uploaded_img(self, file_path):
        self.messageText.clear()
        image_uri = QtCore.QUrl(f"file://{file_path}")
        image = QtGui.QImage(QtGui.QImageReader(file_path).read())
        self.messageText.document().addResource(
            QtGui.QTextDocument.ImageResource,
            image_uri,
            QtCore.QVariant(image)
        )
        img_width, img_height = self.get_img_dimensions(image)
        image_format = QtGui.QTextImageFormat()
        image_format.setWidth(img_width)
        image_format.setHeight(img_height)
        image_format.setName(image_uri.toString())
        cursor = self.messageText.textCursor()
        cursor.movePosition(
            QtGui.QTextCursor.End,
            QtGui.QTextCursor.MoveAnchor
        )
        cursor.insertImage(image_format)
        self.messageText.setReadOnly(True)

    def display_uploaded_document(self, file_path):
        self.messageText.clear()
        self.messageText.setText(f"File Selected: {file_path}")
        self.messageText.setReadOnly(True)

    # Signal Slots
    def handle_search_text_changed(self):
        self.current_search_index = None

    def handle_prev_message_search(self):
        if self.current_search_index is None:
            self.handle_message_search(self.messageList.count() - 1, -1, -1)
        else:
            self.handle_message_search(self.current_search_index - 1, -1, -1)

    def handle_next_message_search(self):
        if self.current_search_index is None:
            self.handle_message_search(self.messageList.count() - 1, -1, -1)
        else:
            self.handle_message_search(self.current_search_index + 1, self.messageList.count())

    def handle_message_search(self, start_index, end_index, step=1):
        text = self.searchMessageLineEdit.text()
        if text == "":
            return
        for row in range(start_index, end_index, step):
            item = self.messageList.item(row)
            message_widget = self.messageList.itemWidget(item)
            if message_widget.message_type in (MessageType.TEXT, MessageType.DOCUMENT):
                if text.lower() in message_widget.message_text.lower():
                    self.messageList.scrollToItem(item, QtWidgets.QAbstractItemView.PositionAtCenter)
                    item.setSelected(True)
                    self.current_search_index = row
                    return
        if self.current_search_index is None:
            show_popup(self, "Alert", "No message found!", 1)

    def toggle_preview(self):
        # Go Back to text box
        if self.messageText.isReadOnly():
            self.messageText.setHtml("")
            self.messageText.setText(self.text)
            self.messageText.moveCursor(QtGui.QTextCursor.End)
            self.messageText.setReadOnly(False)
            self.messageText.setStyleSheet("")
            self.previewBtn.setDefault(False)
            self.previewBtn.setText("Preview")
        # Show preview
        else:
            self.text = self.messageText.toPlainText()
            html = self.markdown.convert(self.text)
            self.messageText.setHtml(html)
            self.messageText.setReadOnly(True)
            self.messageText.setStyleSheet("background: #eff0f1")
            self.previewBtn.setDefault(True)
            self.previewBtn.setText("Write")

    def handle_upload(self):
        img_type = "Image (*.png *.gif *.jpg *jpeg *.bmp)"
        doc_type = "Document (*.*)"
        file_filter = f'{img_type};;{doc_type}'
        file_path = get_open_filename(self, "Select a file", "", file_filter)
        if file_path is None or file_path == "":
            return
        file_type = file_path.split('.')[-1]
        self.attachment = file_path
        if file_type in ['png', 'gif', 'jpg', 'jpeg', 'bmp']:
            self.attachment_type = MessageType.IMAGE
            self.display_uploaded_img(file_path)
        else:
            self.attachment_type = MessageType.DOCUMENT
            self.display_uploaded_document(file_path)
        self.uploadBtn.setVisible(False)
        self.cancelBtn.setVisible(True)
        self.previewBtn.setVisible(False)

    def send_message(self):
        """
        send message through connection
        """
        if self.attachment is None:
            reply_id = -1
            if self.active_message_reply is not None:
                reply_id = self.active_message_reply.id
            message_text = self.messageText.toPlainText()
            if message_text == "":
                return
            message_text = message_text.strip()
            self.conn.send_message(message_text, self.op_id, reply_id)
        else:
            files = {"file": open(self.attachment, 'rb')}
            data = {
                "token": self.token,
                "op_id": self.op_id,
                "message_type": int(self.attachment_type)
            }
            url = url_join(self.mscolab_server_url, 'message_attachment')
            try:
                requests.post(url, data=data, files=files)
            except requests.exceptions.ConnectionError:
                show_popup(self, "Error", "File size too large")
        self.send_message_state()

    def start_message_reply(self, message_item):
        self.send_message_state()
        self.active_message_reply = message_item
        self.active_message_reply.set_selected(True)
        self.uploadBtn.setVisible(False)
        self.cancelBtn.setVisible(True)

    def start_message_edit(self, message_text, message_id):
        self.send_message_state()
        self.active_edit_id = message_id
        self.messageText.setText(message_text)
        self.messageText.setFocus()
        self.messageText.moveCursor(QtGui.QTextCursor.End)
        self.editMessageBtn.setVisible(True)
        self.cancelBtn.setVisible(True)
        self.sendMessageBtn.setVisible(False)
        self.uploadBtn.setVisible(False)

    def send_message_state(self):
        self.active_edit_id = None
        self.attachment = None
        if self.active_message_reply is not None:
            self.active_message_reply.set_selected(False)
            self.active_message_reply = None
        self.messageText.clear()
        self.messageText.setReadOnly(False)
        self.messageText.setFocus()
        self.editMessageBtn.setVisible(False)
        self.cancelBtn.setVisible(False)
        self.sendMessageBtn.setVisible(True)
        self.previewBtn.setVisible(True)
        self.uploadBtn.setVisible(True)

    def edit_message(self):
        new_message_text = self.messageText.toPlainText()
        if new_message_text == "":
            self.conn.delete_message(self.active_edit_id, self.op_id)
        else:
            new_message_text = new_message_text.strip()
            self.conn.edit_message(self.active_edit_id, new_message_text, self.op_id)
        self.send_message_state()

    # API REQUESTS
    def load_users(self):
        # load users to side-tab here
        # make request to get users
        data = {
            "token": self.token,
            "op_id": self.op_id
        }
        url = url_join(self.mscolab_server_url, 'authorized_users')
        r = requests.get(url, data=data)
        if r.text != "False":
            self.collaboratorsList.clear()
            users = r.json()["users"]
            for user in users:
                item = QtWidgets.QListWidgetItem(f'{user["username"]} - {user["access_level"]}',
                                                 parent=self.collaboratorsList)
                self.collaboratorsList.addItem(item)
        else:
            show_popup(self, "Error", "Session expired, new login required")

    def load_all_messages(self):
        # empty messages and reload from server
        data = {
            "token": self.token,
            "op_id": self.op_id,
            "timestamp": datetime.datetime(1970, 1, 1).strftime("%Y-%m-%d, %H:%M:%S")
        }
        # returns an array of messages
        url = url_join(self.mscolab_server_url, "messages")

        res = requests.get(url, data=data)
        if res.text != "False":
            res = res.json()
            messages = res["messages"]
            # clear message box
            for message in messages:
                self.render_new_message(message, scroll=False)
            self.messageList.scrollToBottom()
        else:
            show_popup(self, "Error", "Session expired, new login required")

    def render_new_message(self, message, scroll=True):
        message_item = MessageItem(message, self)
        list_widget_item = QtWidgets.QListWidgetItem(self.messageList)
        list_widget_item.setSizeHint(message_item.sizeHint())
        self.messageList.addItem(list_widget_item)
        self.messageList.setItemWidget(list_widget_item, message_item)
        if scroll:
            self.messageList.scrollToBottom()

    # SOCKET HANDLERS
    @QtCore.Slot(int)
    def handle_permissions_updated(self, _):
        self.load_users()

    @QtCore.Slot(str)
    def handle_incoming_message(self, message):
        message = json.loads(message)
        self.render_new_message(message)

    @QtCore.Slot(str)
    def handle_incoming_message_reply(self, reply):
        reply = json.loads(reply)
        for i in range(self.messageList.count() - 1, -1, -1):
            item = self.messageList.item(i)
            message_widget = self.messageList.itemWidget(item)
            if message_widget.id == reply["reply_id"]:
                # TODO: Hacky Approach. Add UI update function in the widget later instead of creating a new widget
                message_widget.replies.append(reply)
                message = {
                    "id": message_widget.id,
                    "u_id": message_widget.u_id,
                    "username": message_widget.username,
                    "replies": message_widget.replies,
                    "message_type": message_widget.message_type,
                    "time": message_widget.time
                }
                if message_widget.message_type in (MessageType.TEXT, MessageType.SYSTEM_MESSAGE):
                    message["text"] = message_widget.message_text
                else:
                    message["text"] = message_widget.attachment_path
                new_message_item = MessageItem(message, self)
                item.setSizeHint(new_message_item.sizeHint())
                self.messageList.setItemWidget(item, new_message_item)
                break

    @QtCore.Slot(str)
    def handle_message_edited(self, message):
        message = json.loads(message)
        message_id = message["message_id"]
        new_message_text = message["new_message_text"]
        # Loop backwards because it's more likely the message is new than old
        for i in range(self.messageList.count() - 1, -1, -1):
            item = self.messageList.item(i)
            message_widget = self.messageList.itemWidget(item)
            if message_widget.id == message_id:
                message_widget.update_text(new_message_text)
                item.setSizeHint(message_widget.sizeHint())
                break

    @QtCore.Slot(str)
    def handle_deleted_message(self, message):
        message = json.loads(message)
        message_id = message["message_id"]
        # Loop backwards because it's more likely the message is new than old
        for i in range(self.messageList.count() - 1, -1, -1):
            item = self.messageList.item(i)
            message_widget = self.messageList.itemWidget(item)
            if message_widget.id == message_id:
                self.messageList.takeItem(i)
                break

    def closeEvent(self, event):
        self.viewCloses.emit()


class MessageItem(QtWidgets.QWidget):
    def __init__(self, message, chat_window):
        super(MessageItem, self).__init__()
        self.id = message["id"]
        self.u_id = message["u_id"]
        self.username = message["username"]
        self.message_type = message["message_type"]
        self.replies = message["replies"]
        self.time = message["time"]
        self.chat_window = chat_window
        self.message_text = None
        self.attachment_path = None
        self.message_image = None
        self.messageBox = None
        if self.message_type in (MessageType.TEXT, MessageType.SYSTEM_MESSAGE):
            self.message_text = message["text"]
        else:
            self.attachment_path = message["text"]
        self.context_menu = QtWidgets.QMenu(self)
        self.textArea = QtWidgets.QWidget()
        self.replyArea = None
        self.replyScroll = QtWidgets.QScrollArea()
        self.setup_message_box()
        self.setup_message_box_layout()
        self.setup_context_menu()

    def setup_image_message_box(self):
        MAX_WIDTH = MAX_HEIGHT = 300
        self.messageBox = QtWidgets.QLabel()
        if '\\' in self.attachment_path:
            img_url = url_join(self.chat_window.mscolab_server_url,
                               self.attachment_path.replace('\\', '/').split('colabdata')[1])
        else:
            img_url = url_join(self.chat_window.mscolab_server_url, self.attachment_path)
        data = requests.get(img_url).content
        image = QtGui.QImage()
        image.loadFromData(data)
        self.message_image = image
        width, height = image.size().width(), image.size().height()
        if width > height and width > MAX_WIDTH:
            image = image.scaledToWidth(MAX_WIDTH)
        elif height > width and height > MAX_HEIGHT:
            image = image.scaledToHeight(MAX_HEIGHT)
        self.messageBox.setPixmap(QtGui.QPixmap(image))
        self.messageBox.setContentsMargins(0, 5, 0, 5)
        self.messageBox.show()

    def get_text_browser(self, text):
        text_browser = QtWidgets.QTextBrowser()
        html = self.chat_window.markdown.convert(text)
        text_browser.setHtml(html)
        text_browser.setOpenLinks(False)
        text_browser.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        text_browser.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        text_browser.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        text_browser.setAttribute(103)
        text_browser.anchorClicked.connect(self.on_link_click)
        text_browser.show()
        text_browser.setFixedHeight(
            int(text_browser.document().size().height() + text_browser.contentsMargins().top() * 2)
        )
        return text_browser

    def setup_text_message_box(self):
        if self.message_type == MessageType.DOCUMENT:
            doc_url = url_join(self.chat_window.mscolab_server_url, self.attachment_path)
            file_name = fs.path.basename(self.attachment_path)
            self.message_text = f"Document: [{file_name}]({doc_url})"
        self.messageBox = self.get_text_browser(self.message_text)

    def setup_message_box(self):
        if self.message_type == MessageType.IMAGE:
            self.setup_image_message_box()
        else:
            self.setup_text_message_box()
        self.messageBox.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.messageBox.customContextMenuRequested.connect(self.open_context_menu)

    def setup_message_box_layout(self):
        container_layout = QtWidgets.QHBoxLayout()
        text_area_layout = QtWidgets.QVBoxLayout()
        if self.chat_window.user["username"] == self.username:
            text_area_layout.addWidget(self.messageBox)
            self.textArea.setLayout(text_area_layout)
            container_layout.addStretch()
            container_layout.addWidget(self.textArea)
        else:
            username_label = QtWidgets.QLabel(f"{self.username}")
            username_label.setContentsMargins(5, 5, 5, 0)
            label_font = QtGui.QFont()
            label_font.setBold(True)
            username_label.setFont(label_font)
            text_area_layout.addWidget(username_label)
            text_area_layout.addWidget(self.messageBox)
            self.textArea.setLayout(text_area_layout)
            container_layout.addWidget(self.textArea)
            container_layout.addStretch()
        for reply in self.replies:
            self.add_message_reply(reply)
        self.textArea.layout().setSpacing(0)
        self.textArea.layout().setContentsMargins(0, 0, 0, 0)
        container_layout.setSpacing(0)
        container_layout.setContentsMargins(5, 5, 5, 5)
        self.set_message_style()
        self.setLayout(container_layout)

    def set_message_style(self, selected=False):
        sent_msg_color = "#dcf8c6"
        recv_msg_color = "#eff0f1"
        system_msg_color = "#a9d3e0"
        message_selected = "#70bbed"
        if self.chat_window.user["username"] == self.username:
            border = "border-top-left-radius: 5px; border-bottom-left-radius: 5px; border-bottom-right-radius: 5px"
            color = sent_msg_color
        else:
            border = "border-top-right-radius: 5px; border-bottom-left-radius: 5px; border-bottom-right-radius: 5px"
            color = recv_msg_color
        if self.message_type == MessageType.SYSTEM_MESSAGE:
            color = system_msg_color
        if selected is True:
            color = message_selected
        self.messageBox.setStyleSheet("background: transparent; border: none;")
        self.textArea.setStyleSheet(f"background: {color}; {border}")

    def insert_reply_area(self):
        self.replyArea = QtWidgets.QGroupBox()
        reply_area_layout = QtWidgets.QFormLayout()
        reply_area_layout.setSpacing(0)
        reply_area_layout.setContentsMargins(5, 0, 0, 0)
        self.replyArea.setLayout(reply_area_layout)
        self.replyScroll.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.replyScroll.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.replyScroll.setMaximumHeight(150)
        self.replyScroll.setWidgetResizable(True)
        self.replyScroll.setWidget(self.replyArea)
        self.replyScroll.setContentsMargins(0, 0, 0, 0)
        self.textArea.layout().addWidget(self.replyScroll)
        if self.username == self.chat_window.user["username"]:
            color = "#c3f39e"
        else:
            color = "#e2e3e5"
        self.messageBox.setStyleSheet("background: transparent; border-bottom: 1px solid black;")
        self.replyScroll.setStyleSheet(f"background: {color}")

    def add_message_reply(self, reply):
        if self.replyArea is None:
            self.insert_reply_area()
        reply_username_label = QtWidgets.QLabel(f'{reply["username"]}:')
        label_font = QtGui.QFont()
        label_font.setBold(True)
        reply_username_label.setFont(label_font)
        reply_message_box = self.get_text_browser(reply["text"])
        self.replyArea.layout().addRow(reply_username_label, reply_message_box)

    def setup_context_menu(self):
        download_action = self.context_menu.addAction("Download")
        copy_action = self.context_menu.addAction("Copy")
        reply_action = self.context_menu.addAction("Reply")
        edit_action = self.context_menu.addAction("Edit")
        delete_action = self.context_menu.addAction("Delete")
        download_action.triggered.connect(self.handle_download_action)
        copy_action.triggered.connect(self.handle_copy_action)
        reply_action.triggered.connect(self.handle_reply_action)
        edit_action.triggered.connect(self.handle_edit_action)
        delete_action.triggered.connect(self.handle_delete_action)
        # Enabling/Disabling actions based on who sent the message
        if self.username != self.chat_window.user["username"]:
            edit_action.setVisible(False)
            delete_action.setVisible(False)
        # Enabling/Disabling actions based on the type of message
        if self.message_type == MessageType.SYSTEM_MESSAGE:
            download_action.setVisible(False)
            copy_action.setVisible(False)
            reply_action.setVisible(False)
            edit_action.setVisible(False)
            delete_action.setVisible(False)
        elif self.message_type == MessageType.TEXT:
            download_action.setVisible(False)
        elif self.message_type == MessageType.IMAGE or self.message_type == MessageType.DOCUMENT:
            copy_action.setVisible(False)
            edit_action.setVisible(False)

    def open_context_menu(self, pos):
        self.context_menu.exec_(self.messageBox.mapToGlobal(pos))

    def handle_copy_action(self):
        QtWidgets.QApplication.clipboard().setText(self.message_text)

    def handle_download_action(self):
        file_name = fs.path.basename(self.attachment_path)
        file_name, file_ext = fs.path.splitext(file_name)
        # fs.file_picker cannot take filenames that contain dots
        default_filename = file_name.replace('.', '_') + file_ext
        if self.message_type == MessageType.DOCUMENT:
            file_path = get_save_filename(self, "Save Document", default_filename, f"Document (*{file_ext})")
            if file_path is not None:
                file_content = requests.get(url_join(self.chat_window.mscolab_server_url, self.attachment_path)).content
                with open(file_path, "wb") as f:
                    f.write(file_content)
        else:
            file_path = get_save_filename(self, "Save Image", default_filename, f"Image (*{file_ext})")
            if file_path is not None:
                self.message_image.save(file_path)

    def handle_reply_action(self):
        self.chat_window.start_message_reply(self)

    def handle_edit_action(self):
        self.chat_window.start_message_edit(self.message_text, self.id)

    def handle_delete_action(self):
        # disable edit message section if it's active before deleting
        self.chat_window.send_message_state()
        self.chat_window.conn.delete_message(self.id, self.chat_window.op_id)

    def update_text(self, message_text):
        self.message_text = message_text
        html = self.chat_window.markdown.convert(self.message_text)
        self.messageBox.setHtml(html)
        self.messageBox.setFixedHeight(
            int(self.messageBox.document().size().height() + self.messageBox.contentsMargins().top() * 2)
        )
        self.textArea.adjustSize()

    def set_selected(self, selected):
        if selected is True:
            self.set_message_style(selected=True)
        else:
            self.set_message_style()

    def on_link_click(self, url):
        if url.scheme() == "":
            url.setScheme("http")
        QtGui.QDesktopServices.openUrl(url)


# Deregister all the syntax that we don't want to allow
# Can't find any part in documentation where all the syntax names are mentioned
# For future reference to syntax name refer to:
# https://github.com/Python-Markdown/markdown/blob/a06659b62209de98cbc23715addb2b768a245788/markdown/core.py#L100
class DeregisterSyntax(Extension):
    """
    Current Supported syntax:
    *text* : emphasis
    **text** : bold
    - text : unordered list
    1. text : ordered list
    # text : Heading
    [text](link) : link
    <link> : clickable link
    """

    def extendMarkdown(self, md):
        # Deregister block syntax
        md.parser.blockprocessors.deregister('setextheader')
        md.parser.blockprocessors.deregister('hr')
        md.parser.blockprocessors.deregister('quote')

        # Deregister inline syntax
        md.inlinePatterns.deregister('reference')
        md.inlinePatterns.deregister('image_link')
        md.inlinePatterns.deregister('image_reference')
        md.inlinePatterns.deregister('short_reference')
        md.inlinePatterns.deregister('automail')
        md.inlinePatterns.deregister('linebreak')
        md.inlinePatterns.deregister('html')
        md.inlinePatterns.deregister('entity')



# -*- coding: utf-8 -*-
"""

    mslib.msui._tests.test_updater
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to tests msui.updater

    This file is part of MSS.

    :copyright: Copyright 2021 May Bär
    :copyright: Copyright 2021-2022 by the MSS team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import sys
import mock
from PyQt5 import QtWidgets, QtTest

from mslib.msui.updater import UpdaterUI, Updater
from mslib.utils.qt import Worker


def no_conda(args=None, **named_args):
    raise FileNotFoundError


class SubprocessDifferentVersionMock:
    def __init__(self, args=None, **named_args):
        self.returncode = 0
        self.args = args
        if args and "list" in args and "mss" in args:
            self.stdout = "*mss 0.0.0\n"
        else:
            self.stdout = "*mss 999.999.999\n"


class SubprocessSameMock:
    def __init__(self, args=None, **named_args):
        self.stdout = "*mss 999.999.999\n"
        self.returncode = 0
        self.args = args


def create_mock(function, on_success=None, on_failure=None, start=True):
    worker = Worker(function)
    if on_success:
        worker.finished.connect(on_success)
    if on_failure:
        worker.failed.connect(on_failure)
    if start:
        worker.run()
    return worker


class Test_MSS_ShortcutDialog:
    def setup(self):
        self.updater = Updater()
        self.status = ""
        self.update_available = False
        self.update_finished = False

        def update_signal(old, new):
            self.update_available = True

        def update_finished_signal():
            self.update_finished = True

        def status_signal(s):
            self.status = s

        self.updater.on_update_available.connect(update_signal)
        self.updater.on_status_update.connect(status_signal)
        self.updater.on_update_finished.connect(update_finished_signal)
        self.application = QtWidgets.QApplication(sys.argv)

    def teardown(self):
        self.application.quit()
        QtWidgets.QApplication.processEvents()

    @mock.patch("subprocess.Popen", new=SubprocessDifferentVersionMock)
    @mock.patch("subprocess.run", new=SubprocessDifferentVersionMock)
    @mock.patch("mslib.utils.qt.Worker.create", create_mock)
    def test_update_recognised(self):
        self.updater.run()

        assert self.updater.new_version == "999.999.999"
        assert self.update_available
        self.updater.new_version = "0.0.0"

        self.updater.update_mss()
        assert self.status == "Update successful. Please restart MSS."
        assert self.update_finished

    @mock.patch("subprocess.Popen", new=SubprocessSameMock)
    @mock.patch("subprocess.run", new=SubprocessSameMock)
    @mock.patch("mslib.utils.qt.Worker.create", create_mock)
    def test_no_update(self):
        self.updater.run()
        assert self.status == "Your MSS is up to date."
        assert not self.update_available
        assert not self.update_finished

    @mock.patch("subprocess.Popen", new=SubprocessDifferentVersionMock)
    @mock.patch("subprocess.run", new=SubprocessDifferentVersionMock)
    @mock.patch("mslib.utils.qt.Worker.create", create_mock)
    def test_update_failed(self):
        self.updater.run()
        assert self.updater.new_version == "999.999.999"
        assert self.update_available
        self.updater.new_version = "1000.1000.1000"
        self.updater.update_mss()
        assert self.status == "Update failed. Please try it manually or " \
                              "by creating a new environment!"

    @mock.patch("subprocess.Popen", new=no_conda)
    @mock.patch("subprocess.run", new=no_conda)
    @mock.patch("mslib.utils.qt.Worker.create", create_mock)
    def test_no_conda(self):
        self.updater.run()
        assert self.updater.new_version is None and self.updater.old_version is None
        assert not self.update_available
        assert not self.update_finished

    @mock.patch("subprocess.Popen", new=no_conda)
    @mock.patch("subprocess.run", new=no_conda)
    @mock.patch("mslib.utils.qt.Worker.create", create_mock)
    def test_exception(self):
        self.updater.new_version = "999.999.999"
        self.updater.old_version = "999.999.999"
        self.updater.update_mss()
        assert self.status == "Update failed, please do it manually."
        assert not self.update_finished

    @mock.patch("subprocess.Popen", new=SubprocessSameMock)
    @mock.patch("subprocess.run", new=SubprocessSameMock)
    @mock.patch("PyQt5.QtWidgets.QMessageBox.information", return_value=QtWidgets.QMessageBox.Yes)
    @mock.patch("mslib.utils.qt.Worker.create", create_mock)
    def test_ui(self, mock):
        ui = UpdaterUI()
        ui.updater.on_update_available.emit("", "")
        QtTest.QTest.qWait(100)
        assert ui.statusLabel.text() == "Update successful. Please restart MSS."
        assert ui.btRestart.isEnabled()



# -*- coding: utf-8 -*-
"""

    mslib.utils._tests.test_coordinate
    ~~~~~~~~~~~~~~~~~~~~~~~

    This module provides pytest functions to tests mslib.utils.coordinate

    This file is part of MSS.

    :copyright: Copyright 2016-2017 Reimar Bauer
    :copyright: Copyright 2016-2022 by the MSS team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import logging
import datetime

import numpy as np
import pytest

import mslib.utils.coordinate as coordinate

LOGGER = logging.getLogger(__name__)


class TestGetDistance(object):
    """
    tests for distance based calculations
    """
    # we don't test the utils method here, may be that method should me refactored off

    def test_get_distance(self):
        coordinates_distance = [(50.355136, 7.566077, 50.353968, 4.577915, 212),
                                (-5.135943, -42.792442, 4.606085, 120.028077, 18130)]
        for lat0, lon0, lat1, lon1, distance in coordinates_distance:
            assert int(coordinate.get_distance(lat0, lon0, lat1, lon1)) == distance

    def test_find_location(self):
        assert coordinate.find_location(50.92, 6.36) == ([50.92, 6.36], 'Juelich')
        assert coordinate.find_location(50.9200002, 6.36) == ([50.92, 6.36], 'Juelich')


class TestProjections(object):
    def test_get_projection_params(self):
        assert coordinate.get_projection_params("epsg:4839") == {'basemap': {'epsg': '4839'}, 'bbox': 'meter(10.5,51)'}
        with pytest.raises(ValueError):
            coordinate.get_projection_params('auto2:42005')
        with pytest.raises(ValueError):
            coordinate.get_projection_params('auto:42001')
        with pytest.raises(ValueError):
            coordinate.get_projection_params('crs:84')


class TestAngles(object):
    """
    tests about angles
    """

    def test_normalize_angle(self):
        assert coordinate.fix_angle(0) == 0
        assert coordinate.fix_angle(180) == 180
        assert coordinate.fix_angle(270) == 270
        assert coordinate.fix_angle(-90) == 270
        assert coordinate.fix_angle(-180) == 180
        assert coordinate.fix_angle(-181) == 179
        assert coordinate.fix_angle(420) == 60

    def test_rotate_point(self):
        assert coordinate.rotate_point([0, 0], 0) == (0.0, 0.0)
        assert coordinate.rotate_point([0, 0], 180) == (0.0, 0.0)
        assert coordinate.rotate_point([1, 0], 0) == (1.0, 0.0)
        assert coordinate.rotate_point([100, 90], 90) == (-90, 100)


class TestLatLonPoints(object):
    def test_linear(self):
        ref_lats = [0, 10]
        ref_lons = [0, 0]

        lats, lons = coordinate.latlon_points(ref_lats[0], ref_lons[0], ref_lats[1], ref_lons[1],
                                              numpoints=2, connection="linear")
        assert len(lats) == len(ref_lats)
        assert all(lats == ref_lats)
        assert len(lons) == len(ref_lons)
        assert all(lons == ref_lons)

        lats, lons = coordinate.latlon_points(ref_lats[0], ref_lons[0], ref_lats[1], ref_lons[1],
                                              numpoints=3, connection="linear")
        assert len(lats) == 3
        assert len(lons) == 3
        assert all(lats == [0, 5, 10])

        ref_lats = [0, 0]
        ref_lons = [0, 10]
        lats, lons = coordinate.latlon_points(ref_lats[0], ref_lons[0], ref_lats[1], ref_lons[1],
                                              numpoints=3, connection="linear")
        assert len(lats) == 3
        assert len(lons) == 3
        assert all(lons == [0, 5, 10])

        lats, lons = coordinate.latlon_points(ref_lats[0], ref_lons[0], ref_lats[1], ref_lons[1],
                                              numpoints=3, connection="linear")
        assert len(lats) == 3
        assert len(lons) == 3
        assert all(lons == [0, 5, 10])

    def test_greatcircle(self):
        ref_lats = [0, 10]
        ref_lons = [0, 0]

        lats, lons = coordinate.latlon_points(ref_lats[0], ref_lons[0], ref_lats[1], ref_lons[1],
                                              numpoints=2, connection="greatcircle")
        assert len(lats) == len(ref_lats)
        assert lats == ref_lats
        assert len(lons) == len(ref_lons)
        assert lons == ref_lons

        lats, lons = coordinate.latlon_points(ref_lats[0], ref_lons[0], ref_lats[1], ref_lons[1],
                                              numpoints=3, connection="linear")
        assert len(lats) == 3
        assert len(lons) == 3
        assert all(np.asarray(lats) == [0, 5, 10])

        ref_lats = [0, 0]
        ref_lons = [0, 10]
        lats, lons = coordinate.latlon_points(ref_lats[0], ref_lons[0], ref_lats[1], ref_lons[1],
                                              numpoints=3, connection="linear")
        assert len(lats) == 3
        assert len(lons) == 3
        assert all(np.asarray(lons) == [0, 5, 10])


def test_pathpoints():
    lats = [0, 10]
    lons = [0, 10]
    times = [datetime.datetime(2012, 7, 1, 10, 30),
             datetime.datetime(2012, 7, 1, 10, 40)]
    ref = [lats, lons, times]
    result = coordinate.path_points(lats, lons, 100, times=times, connection="linear")
    assert all(len(_x) == 100 for _x in result)
    for i in range(3):
        assert pytest.approx(result[i][0]) == ref[i][0]
        assert pytest.approx(result[i][-1]) == ref[i][-1]

    result = coordinate.path_points(lats, lons, 100, times=times, connection="greatcircle")
    assert all(len(_x) == 100 for _x in result)
    for i in range(3):
        assert pytest.approx(result[i][0]) == ref[i][0]
        assert pytest.approx(result[i][-1]) == ref[i][-1]

    result = coordinate.path_points(lats, lons, 200, times=times, connection="linear")
    assert all(len(_x) == 200 for _x in result)
    for i in range(3):
        assert pytest.approx(result[i][0]) == ref[i][0]
        assert pytest.approx(result[i][-1]) == ref[i][-1]

    result = coordinate.path_points(lats, lons, 200, times=times, connection="greatcircle")
    assert all(len(_x) == 200 for _x in result)
    for i in range(3):
        assert pytest.approx(result[i][0]) == ref[i][0]
        assert pytest.approx(result[i][-1]) == ref[i][-1]

    lats = [0, 10, -20]
    lons = [0, 10, 20]
    times = [datetime.datetime(2012, 7, 1, 10, 30),
             datetime.datetime(2012, 7, 1, 10, 40),
             datetime.datetime(2012, 7, 1, 10, 50)]
    ref = [lats, lons, times]

    result = coordinate.path_points(lats, lons, 100, times=times, connection="linear")
    assert all([len(_x) == 100 for _x in result])
    for i in range(3):
        assert pytest.approx(result[i][0]) == ref[i][0]
        assert pytest.approx(result[i][-1]) == ref[i][-1]

    result = coordinate.path_points(lats, lons, 100, times=times, connection="greatcircle")
    assert all(len(_x) == 100 for _x in result)
    for i in range(3):
        assert pytest.approx(result[i][0]) == ref[i][0]
        assert pytest.approx(result[i][-1]) == ref[i][-1]



# -*- coding: utf-8 -*-
"""

    mslib.msui._tests.test_mpl_map
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module is used to test the functionality of mpl_map.py

    This file is part of MSS.

    :copyright: Copyright 2021 May Bär
    :copyright: Copyright 2021-2022 by the MSS team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from matplotlib import pyplot as plt
from mslib.msui.mpl_map import MapCanvas


class Test_MapCanvas:
    def setup(self):
        kwargs = {'resolution': 'l', 'area_thresh': 1000.0, 'ax': plt.gca(), 'llcrnrlon': -15.0, 'llcrnrlat': 35.0,
                  'urcrnrlon': 30.0, 'urcrnrlat': 65.0, 'epsg': '4326'}
        self.map = MapCanvas(**kwargs)

    def test_no_coastsegs(self):
        """
        Assert rendering areas without a coastline does not cause an error
        """
        # On land
        self.map.ax.set_xlim([2, 3])
        self.map.ax.set_ylim([48, 49])
        self.map._draw_auto_graticule()
        self.map.update_with_coordinate_change()

        # On water
        self.map.ax.set_xlim([1, 2])
        self.map.ax.set_ylim([62, 63])
        self.map.update_with_coordinate_change()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_hexagon_dockwidget.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_HexagonDockWidget(object):
    def setupUi(self, HexagonDockWidget):
        HexagonDockWidget.setObjectName("HexagonDockWidget")
        HexagonDockWidget.resize(638, 114)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(HexagonDockWidget.sizePolicy().hasHeightForWidth())
        HexagonDockWidget.setSizePolicy(sizePolicy)
        self.verticalLayout = QtWidgets.QVBoxLayout(HexagonDockWidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label = QtWidgets.QLabel(HexagonDockWidget)
        self.label.setObjectName("label")
        self.horizontalLayout_5.addWidget(self.label)
        self.dsbHexagonLatitude = QtWidgets.QDoubleSpinBox(HexagonDockWidget)
        self.dsbHexagonLatitude.setMinimumSize(QtCore.QSize(0, 0))
        self.dsbHexagonLatitude.setPrefix("")
        self.dsbHexagonLatitude.setMinimum(-90.0)
        self.dsbHexagonLatitude.setMaximum(90.0)
        self.dsbHexagonLatitude.setObjectName("dsbHexagonLatitude")
        self.horizontalLayout_5.addWidget(self.dsbHexagonLatitude)
        self.label_2 = QtWidgets.QLabel(HexagonDockWidget)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_5.addWidget(self.label_2)
        self.dsbHexagonLongitude = QtWidgets.QDoubleSpinBox(HexagonDockWidget)
        self.dsbHexagonLongitude.setMinimum(-180.0)
        self.dsbHexagonLongitude.setMaximum(360.0)
        self.dsbHexagonLongitude.setObjectName("dsbHexagonLongitude")
        self.horizontalLayout_5.addWidget(self.dsbHexagonLongitude)
        self.verticalLayout.addLayout(self.horizontalLayout_5)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_3 = QtWidgets.QLabel(HexagonDockWidget)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_2.addWidget(self.label_3)
        self.dsbHexgaonRadius = QtWidgets.QDoubleSpinBox(HexagonDockWidget)
        self.dsbHexgaonRadius.setMinimum(10.0)
        self.dsbHexgaonRadius.setMaximum(9999.99)
        self.dsbHexgaonRadius.setSingleStep(10.0)
        self.dsbHexgaonRadius.setProperty("value", 200.0)
        self.dsbHexgaonRadius.setObjectName("dsbHexgaonRadius")
        self.horizontalLayout_2.addWidget(self.dsbHexgaonRadius)
        self.label_4 = QtWidgets.QLabel(HexagonDockWidget)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_2.addWidget(self.label_4)
        self.dsbHexagonAngle = QtWidgets.QDoubleSpinBox(HexagonDockWidget)
        self.dsbHexagonAngle.setMinimum(-360.0)
        self.dsbHexagonAngle.setMaximum(360.0)
        self.dsbHexagonAngle.setObjectName("dsbHexagonAngle")
        self.horizontalLayout_2.addWidget(self.dsbHexagonAngle)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.pbAddHexagon = QtWidgets.QPushButton(HexagonDockWidget)
        self.pbAddHexagon.setObjectName("pbAddHexagon")
        self.horizontalLayout_6.addWidget(self.pbAddHexagon)
        self.pbRemoveHexagon = QtWidgets.QPushButton(HexagonDockWidget)
        self.pbRemoveHexagon.setObjectName("pbRemoveHexagon")
        self.horizontalLayout_6.addWidget(self.pbRemoveHexagon)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_6.addItem(spacerItem)
        self.cbClock = QtWidgets.QComboBox(HexagonDockWidget)
        self.cbClock.setMinimumSize(QtCore.QSize(120, 0))
        self.cbClock.setObjectName("cbClock")
        self.cbClock.addItem("")
        self.cbClock.addItem("")
        self.horizontalLayout_6.addWidget(self.cbClock)
        self.verticalLayout.addLayout(self.horizontalLayout_6)

        self.retranslateUi(HexagonDockWidget)
        QtCore.QMetaObject.connectSlotsByName(HexagonDockWidget)

    def retranslateUi(self, HexagonDockWidget):
        _translate = QtCore.QCoreApplication.translate
        HexagonDockWidget.setWindowTitle(_translate("HexagonDockWidget", "Remote Sensing"))
        self.label.setText(_translate("HexagonDockWidget", "Center latitude"))
        self.dsbHexagonLatitude.setSuffix(_translate("HexagonDockWidget", " °N"))
        self.label_2.setText(_translate("HexagonDockWidget", "Center longitude"))
        self.dsbHexagonLongitude.setSuffix(_translate("HexagonDockWidget", " °E"))
        self.label_3.setText(_translate("HexagonDockWidget", "Radius"))
        self.dsbHexgaonRadius.setSuffix(_translate("HexagonDockWidget", " km"))
        self.label_4.setText(_translate("HexagonDockWidget", "Angle of first point"))
        self.dsbHexagonAngle.setSuffix(_translate("HexagonDockWidget", " °"))
        self.pbAddHexagon.setText(_translate("HexagonDockWidget", "Add Hexagon"))
        self.pbRemoveHexagon.setText(_translate("HexagonDockWidget", "Remove Hexagon"))
        self.cbClock.setItemText(0, _translate("HexagonDockWidget", "clockwise"))
        self.cbClock.setItemText(1, _translate("HexagonDockWidget", "anti-clockwise"))



# (mss)
from PyQt5 import QtCore, QtWidgets

# (mss)
from mslib.support.qt_json_view import delegate
from mslib.support.qt_json_view.datatypes import TypeRole


class JsonView(QtWidgets.QTreeView):
    """Tree to display the JsonModel."""

    def __init__(self, parent=None):
        super(JsonView, self).__init__(parent=parent)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._menu)
        self.setItemDelegate(delegate.JsonDelegate())

    def _menu(self, position):
        """Show the actions of the DataType (if any)."""
        menu = QtWidgets.QMenu()
        index = self.indexAt(position)
        data = index.data(TypeRole)
        if data is None:
            return
        for action in data.actions(index):
            menu.addAction(action)
        menu.exec_(self.viewport().mapToGlobal(position))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_add_user.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_addUserDialog(object):
    def setupUi(self, addUserDialog):
        addUserDialog.setObjectName("addUserDialog")
        addUserDialog.resize(375, 232)
        self.gridLayout = QtWidgets.QGridLayout(addUserDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        self.formLayout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.formLayout.setVerticalSpacing(14)
        self.formLayout.setObjectName("formLayout")
        self.username = QtWidgets.QLineEdit(addUserDialog)
        self.username.setObjectName("username")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.username)
        self.emailid = QtWidgets.QLineEdit(addUserDialog)
        self.emailid.setObjectName("emailid")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.emailid)
        self.passwordLabel = QtWidgets.QLabel(addUserDialog)
        self.passwordLabel.setObjectName("passwordLabel")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.passwordLabel)
        self.password = QtWidgets.QLineEdit(addUserDialog)
        self.password.setEchoMode(QtWidgets.QLineEdit.Password)
        self.password.setObjectName("password")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.password)
        self.confirmPasswordLabel = QtWidgets.QLabel(addUserDialog)
        self.confirmPasswordLabel.setObjectName("confirmPasswordLabel")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.confirmPasswordLabel)
        self.rePassword = QtWidgets.QLineEdit(addUserDialog)
        self.rePassword.setEchoMode(QtWidgets.QLineEdit.Password)
        self.rePassword.setObjectName("rePassword")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.rePassword)
        self.emailIDLabel = QtWidgets.QLabel(addUserDialog)
        self.emailIDLabel.setObjectName("emailIDLabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.emailIDLabel)
        self.usernameLabel = QtWidgets.QLabel(addUserDialog)
        self.usernameLabel.setObjectName("usernameLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.usernameLabel)
        self.verticalLayout.addLayout(self.formLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(addUserDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)
        self.gridLayout.addLayout(self.verticalLayout, 0, 0, 1, 1)

        self.retranslateUi(addUserDialog)
        self.buttonBox.accepted.connect(addUserDialog.accept)
        self.buttonBox.rejected.connect(addUserDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(addUserDialog)

    def retranslateUi(self, addUserDialog):
        _translate = QtCore.QCoreApplication.translate
        addUserDialog.setWindowTitle(_translate("addUserDialog", "Add user"))
        self.username.setPlaceholderText(_translate("addUserDialog", "John Doe"))
        self.emailid.setPlaceholderText(_translate("addUserDialog", "johndoe@gmail.com"))
        self.passwordLabel.setText(_translate("addUserDialog", "Password:"))
        self.password.setPlaceholderText(_translate("addUserDialog", "Your password"))
        self.confirmPasswordLabel.setText(_translate("addUserDialog", "Confirm Password:"))
        self.rePassword.setPlaceholderText(_translate("addUserDialog", "Confirm your password"))
        self.emailIDLabel.setText(_translate("addUserDialog", "Email:"))
        self.usernameLabel.setText(_translate("addUserDialog", "Username:"))

